# -*- coding: utf-8 -*-
"""
Created on Mon Aug 16 09:58:31 2021

@author: Filippo Palomba
"""
# Temporary code to suppress pandas FutureWarning
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import numpy
import pandas
import nlopt
import os
import cvxpy
import shutil
import statsmodels.api as sm
from math import sqrt, log, ceil
from copy import deepcopy
from sklearn.preprocessing import PolynomialFeatures

def complete_cases(x):
    return numpy.all(numpy.invert(numpy.isnan(x)), axis=1)


def crossprod(x, y=None):
    if y is None:
        return x.T @ x
    else:
        return x.T @ y


# Auxiliary functions for estimation
def obj_fun_est(x, grad, Z, V, A):
    Gram = Z.T.dot(V).dot(Z)
    a = 2 * numpy.array(A).T @ numpy.array(V) @ numpy.array(Z)

    if grad.size > 0:
        grad[:] = 2 * Gram.dot(x) - a

    f = x.T.dot(Gram).dot(x) - a.dot(x)

    return f[0]


def norm_co_est(x, grad, p, J, KM, QQ, dire):
    if p == 1:
        if grad.size > 0 and dire == "==":
            grad[:] = [1] * J + [0] * KM
        elif grad.size > 0:
            aux = [1 if xi > 0 else -1 for xi in x[0:J]]
            grad[:] = numpy.append(aux, [0] * KM)
        co = numpy.sum(abs(x[0:J])) - QQ
    else:
        if grad.size > 0:
            grad[:] = numpy.append(2 * x[0:J], [0] * KM)
        co = numpy.sum(x[0:J]**2) - QQ**2

    return co


# Auxiliary functions for inference

def obj_fun_min(x, grad, xt, beta):
    f = -numpy.sum(xt * (x - beta))
    if grad.size > 0:
        grad[:] = -xt
    return f


def obj_fun_max(x, grad, xt, beta):
    f = numpy.sum(xt * (x - beta))
    if grad.size > 0:
        grad[:] = xt
    return f


# Loss function constraint
def single_ineq(x, grad, beta, Q, G):
    a = -2 * G - 2 * beta.T.dot(Q)
    d = 2 * G.dot(beta) + beta.dot(Q.dot(beta))
    f = x.T.dot(Q).dot(x) + a.dot(x) + d
    if grad.size > 0:
        grad[:] = 2 * Q.dot(x) + a

    return f


# Norm constraint
def norm_equal(x, grad, J, KM, QQ, p_int, dire):
    if p_int == 1:
        if grad.size > 0 and dire == "==":
            grad[:] = [1] * J + [0] * KM
        elif grad.size > 0:
            aux = [1 if xi > 0 else -1 for xi in x[0:J]]
            grad[:] = numpy.append(aux, [0] * KM)
        co = numpy.sum(abs(x[0:J])) - QQ
    else:
        if grad.size > 0:
            grad[:] = numpy.append(2 * x[0:J], [0] * KM)
        co = numpy.sum(x[0:J]**2) - QQ**2

    return co


def w_constr_prep(cons, names, A, Z, V, J, KM):
    # Default method to estimate weights as in Abadie et al. (2010)
    if cons is None:
        cons = {'lb': 0,
                'p': 'L1',
                'dir': '==',
                'Q': 1,
                'name': 'simplex'}
    elif cons['name'] == 'simplex':
        if 'Q' not in names:
            cons['Q'] = 1
        cons = {'lb': 0,
                'p': 'L1',
                'dir': '==',
                'Q': cons['Q'],
                'name': 'simplex'}
    elif cons['name'] == 'ols':
        cons = {'lb': -numpy.inf,
                'p': 'no norm',
                'dir': None,
                'Q': None,
                'name': 'ols'}
    elif cons['name'] == 'lasso':
        if 'Q' not in names:
            cons['Q'], lambd = shrinkage_EST("lasso", A, Z, V, J, KM)
        cons = {'lb': -numpy.inf,
                'p': 'L1',
                'dir': '<=',
                'Q': cons['Q'],
                'name': 'lasso'}
    elif cons['name'] == 'ridge':
        Qest, lambd = shrinkage_EST("ridge", A, Z, V, J, KM)
        if 'Q' not in names:
            cons['Q'] = Qest
        cons = {'lb': -numpy.inf,
                'p': 'L2',
                'dir': '<=',
                'Q': cons['Q'],
                'name': 'ridge',
                'lambda': lambd}

    else:
        if not all(req in names for req in ['p', 'dir', 'Q', 'lb']):
            raise Exception("If 'name' is not specified, w_constr should be" +
                            " a list whose elements  must be named 'p','dir','Q','lb'.")
        if cons['p'] not in ['no norm', 'L1', 'L2']:
            raise Exception("In w_constr specify either p = 'no norm' (no constraint on the norm of weights)," +
                            " p = 'L1' (L1-norm), p = 'L2' (L2-norm).")
        if cons['lb'] != 0 and cons['lb'] != -numpy.inf:
            raise Exception("In w_constr specify either lb = 0 or lb = -numpy.inf.")
        cons['name'] = "user provided"
    return cons


def shrinkage_EST(method, A, Z, V, J, KM):
    lamb = None
    if method == "lasso":
        Q = 1

    if method == "ridge":
        wls = sm.WLS(A, Z, weights=numpy.diag(V)).fit()
        L2 = sqrt(sum(wls.params**2))
        Ahat = wls.predict(Z)
        sig = numpy.sum((A.iloc[:, 0] - Ahat)**2) / (len(A) - (J + KM))
        lamb = sig * (J + KM) / L2            # rule of thumb
        Q = L2 / (1 + lamb)                 # convert lambda into Q

    return Q, lamb


def b_est(A, Z, J, KM, w_constr, V, opt_dict):

    lb = w_constr['lb']
    dire = w_constr['dir']
    p = w_constr['p']

    if p == "no norm":
        pp = 0
    if p == "L1":
        pp = 1
    if p == "L2":
        pp = 2

    Zarr = numpy.array(Z)

    opt_dict = prepareOptions(opt_dict, "scest")

    use_cvxpy = useCVXPY(w_constr)

    if use_cvxpy is True:
        x = cvxpy.Variable((J + KM, 1))

        omega_inv = numpy.linalg.pinv(V)
        C = numpy.linalg.cholesky(omega_inv)
        C_inv = numpy.linalg.pinv(C)
        A_star = C_inv.dot(A)
        Z_star = C_inv.dot(Zarr)
        objective = cvxpy.Minimize(cvxpy.sum_squares(Z_star @ x - A_star))
        constraints = [cvxpy.norm1(x[0:J]) <= w_constr['Q'], x[0:J] >= lb]
        prob = cvxpy.Problem(objective, constraints)
        prob.solve()

        b = x.value
        alert = prob.status != 'optimal'

    else:

        if pp == 1 and lb == -numpy.inf:
            opt = nlopt.opt(nlopt.LD_MMA, J + KM)
        else:
            opt = nlopt.opt(nlopt.LD_SLSQP, J + KM)

        opt.set_min_objective(lambda x, grad: obj_fun_est(x, grad, Zarr, V, A))

        if dire == "==":
            opt.add_equality_constraint(lambda x, grad: norm_co_est(x, grad,
                                        pp, J, KM, w_constr['Q'], dire),
                                        opt_dict['tol_eq'])

        elif dire == "<=":
            opt.add_inequality_constraint(lambda x, grad: norm_co_est(x,
                                          grad, pp, J, KM, w_constr['Q'],
                                          dire), opt_dict['tol_ineq'])

        opt.set_lower_bounds([lb] * J + [-numpy.inf] * KM)
        opt.set_xtol_rel(1e-16)
        opt.set_xtol_abs(1e-16)
        opt.set_ftol_rel(1e-16)
        opt.set_ftol_abs(1e-16)
        opt.set_maxeval(opt_dict['maxeval'])
        b = opt.optimize([0] * (J + KM))

        last_res = opt.last_optimize_result()
        alert = last_res < 0 or last_res >= 5
        if alert is True:
            raise Exception("Estimation algorithm not converged! The algorithm returned the value: " +
                            str(opt.last_optimize_result()) + ". To check to what errors it corresponds" +
                            "go to 'https://nlopt.readthedocs.io/en/latest/NLopt_Reference/#return-values'.")

    return b


def u_des_prep(B, C, Z, u_order, u_lags, coig_data, T0_tot, M, constant, index,
               index_w, u_design, res):

    # Construct the polynomial terms in B
    if u_order == 0:       # Simple mean
        u_des_0 = pandas.DataFrame(numpy.ones(T0_tot), index=B.index)

    elif u_order > 0:  # Include covariates when predicting u_mean

        # Create first differences feature-by-feature of the matrix B (not of C!!)
        if coig_data is True:
            B_diff = B - B.groupby('feature').shift(1)
            u_des_0 = pandas.concat([B_diff, C], axis=1).loc[:, index]

        elif coig_data is False:
            u_des_0 = Z.loc[:, index]

        if constant is False:
            u_des_0.insert(loc=len(u_des_0.columns), column='constant_term_u',
                           value=numpy.ones(len(u_des_0)))

    # polynomials of second order are handled later in the code

    # Construct lags of B
    if u_lags > 0:
        B_lags = pandas.DataFrame(None, index=B.index)

        if coig_data is True:
            B_diff = B - B.groupby('feature').shift(1)

        for ll in range(1, u_lags + 1):
            if coig_data is False:
                B_lags = pandas.concat([B_lags,
                                        B.loc[:, index_w].groupby('feature').shift(ll)], axis=1)
            else:
                B_lags = pandas.concat([B_lags,
                                        B_diff.loc[:, index_w].groupby('feature').shift(ll)], axis=1)

        u_des_0 = pandas.concat([u_des_0, B_lags], axis=1)

    # User chose U design matrix
    if u_design is not None:
        if not isinstance(u_design, (pandas.DataFrame, numpy.ndarray)):
            raise Exception("The object u_design should be a dataframe or a matrix!")

        if isinstance(u_design, numpy.ndarray):
            u_design = pandas.DataFrame(u_design,
                                        index=res.index,
                                        columns=res.columns)

        if len(u_design) != len(res):
            raise Exception("The object u_design has " + len(u_design) + " rows when " +
                            len(res) + " where expected!")

        u_des_0 = u_design

    return u_des_0


def e_des_prep(B, C, Z, P, e_order, e_lags, res, sc_pred, Y_donors, out_feat, J, M, index,
               index_w, coig_data, T0, T0_tot, T1, outcome_var, constant, e_design):
    if out_feat is False:
        e_res = sc_pred.Y_pre - sc_pred.Y_pre_fit

        if coig_data is True:
            e_des_0 = Y_donors.loc[:, index_w] - Y_donors.loc[:, index_w].shift(1)
            P_first = P.iloc[[0], :J] - Y_donors.iloc[[len(Y_donors) - 1], :].values
            P_diff = P.iloc[:, :J].diff()
            P_diff.iloc[0, :] = P_first
            e_des_1 = P_diff.loc[:, index_w]
        else:
            e_des_0 = Y_donors.loc[:, index_w]
            e_des_1 = P.loc[:, index_w]

    elif out_feat is True:
        e_res = res.loc[(outcome_var,), ]

        # Construct the polynomial terms in B (e.des.0) and P (e.des.1)
        if e_order == 0:       # Simple mean
            e_des_0 = pandas.DataFrame(numpy.ones(T0_tot), index=B.index)
            e_des_1 = pandas.DataFrame(numpy.ones(T1), index=P.index)

        elif e_order > 0:  # Include covariates when predicting u_mean
            # Create first differences feature-by-feature of the matrix B (not of C!!)
            if coig_data is True:
                B_diff = B - B.groupby('feature').shift(1)
                e_des_0 = pandas.concat([B_diff, C], axis=1).loc[:, index]

                # Remove last observation of first feature from first period of P
                P_first = P.iloc[[0], :J] - B.iloc[[T0[outcome_var] - 1], :].values
                P_diff = P.iloc[:, :J].diff()
                P_diff.iloc[0, :] = P_first
                e_des_1 = pandas.concat([P_diff.loc[:, index_w], P.iloc[:, J:]], axis=1)

            elif coig_data is False:
                e_des_0 = Z.loc[:, index]
                e_des_1 = P.loc[:, index]

            if constant is False:
                e_des_0.insert(loc=len(e_des_0.columns), column='0_constant',
                               value=numpy.ones(len(e_des_0)))
                e_des_1.insert(loc=len(e_des_1.columns), column='0_constant',
                               value=numpy.ones(len(e_des_1)))

        # Construct lags of B
        if e_lags > 0:
            B_lags = pandas.DataFrame(None, index=B.index)
            P_lags = pandas.DataFrame(None, index=P.index)
    
            B_diff = B - B.groupby('feature').shift(1)
            aux = B.iloc[B.index.get_level_values('feature') == outcome_var]
            if coig_data is False:
                df1 = aux.iloc[(len(aux) - e_lags):len(aux)]
                df2 = P.iloc[:, 0:J]
                P_aug = pandas.concat([df1, df2], axis=0)
            else:
                df1 = aux.iloc[(len(aux) - e_lags - 1):len(aux)]
                df2 = P.iloc[:, 0:J]
                P_aug = pandas.concat([df1, df2], axis=0)
    
            P_aug_diff = P_aug - P_aug.shift(1)
    
            for ll in range(1, e_lags + 1):
                if coig_data is False:
                    B_lags = pandas.concat([B_lags,
                                           B.loc[:, index_w].groupby('feature').shift(ll)], axis=1)
                    P_lags = pandas.concat([P_lags,
                                           P_aug.iloc[e_lags:, ].loc[:, index_w].shift(ll)], axis=1)
                else:
                    B_lags = pandas.concat([B_lags,
                                           B_diff.loc[:, index_w].groupby('feature').shift(ll)], axis=1)
                    P_lags = pandas.concat([P_lags,
                                           P_aug_diff.shift(ll).iloc[(e_lags + 1):, ].loc[:, index_w]], axis=1)
    
            e_des_0 = pandas.concat([e_des_0, B_lags], axis=1)
            e_des_1 = pandas.concat([e_des_1, P_lags], axis=1)
    
        e_des_0 = e_des_0.loc[(outcome_var,), ]
    
        if e_design is not None:
            if not isinstance(e_design, (pandas.DataFrame, numpy.ndarray)):
                raise Exception("The object e_design should be a dataframe or a matrix!")
    
            if isinstance(e_design, numpy.ndarray):
                e_design = pandas.DataFrame(e_design,
                                            index=res.index,
                                            columns=res.columns)
    
            if len(e_design) != len(e_res):
                raise Exception("The object e_design has " + len(e_design) + " rows when " +
                                len(e_res) + " where expected!")
    
            e_des_0 = e_design

    return e_res, e_des_0, e_des_1

def createPoly(u_order, e_order, index_w, u_des_0_na, e_des_0_na, e_des_1, out_feat):
    if u_order > 1:
        poly = PolynomialFeatures(u_order, include_bias=False)     # create pipeline
        act_B = len(index_w)                   # count active B columns in u_des_0
        B_poly = poly.fit_transform(u_des_0_na.iloc[:, :act_B])  # evaluate poly with interactions
        B_poly = pandas.DataFrame(B_poly, index=u_des_0_na.index)
        u_des_0_na = pandas.concat([B_poly, u_des_0_na.iloc[:, act_B:]], axis=1)

    if e_order > 1 and out_feat is True:
        poly = PolynomialFeatures(e_order, include_bias=False)     # create pipeline

        # Model e
        act_B = len(index_w)                   # count active B columns in e_des_0
        B_poly = poly.fit_transform(e_des_0_na.iloc[:, 0:act_B])  # evaluate poly with interactions
        B_poly = pandas.DataFrame(B_poly, index=e_des_0_na.index)
        e_des_0_na = pandas.concat([B_poly, e_des_0_na.iloc[:, act_B:]], axis=1)

        # Predictors for e
        act_P = len(index_w)                   # count active B columns in e_des_1
        P_poly = poly.fit_transform(e_des_1.iloc[:, 0:act_P])  # evaluate poly with interactions
        P_poly = pandas.DataFrame(P_poly, index=e_des_1.index)
        e_des_1 = pandas.concat([P_poly, e_des_1.iloc[:, act_P:]], axis=1)

    return u_des_0_na, e_des_0_na, e_des_1

def DUflexGet(u_des_0, C):
    sel_u_des_B = [u not in C.columns.tolist() for u in u_des_0.columns]
    sel_u_des_C = [b is False for b in sel_u_des_B]
    D_b = u_des_0.loc[:, sel_u_des_B]
    D_c = u_des_0.loc[:, sel_u_des_C]
    f_id = pandas.get_dummies(u_des_0.index.get_level_values('feature'))
    f_id.set_index(D_b.index, inplace=True)
    D_bb = pandas.concat([D_b, f_id], axis=1)
    features = f_id.columns.tolist()
    tomult = D_b.columns.tolist()

    D_b_int = pandas.DataFrame(index=D_b.index)
    for f in features:
        aux = D_bb.loc[:, tomult].multiply(D_bb.loc[:, f], axis="index")
        D_b_int = pandas.concat([D_b_int, aux], axis=1)

    D_b_int.set_index(D_c.index, inplace=True)
    Du = pandas.concat([D_b_int, D_c], axis=1)

    return Du

def df_EST(w_constr, w, A, B, J, KM):
    if (w_constr['name'] == "ols") or (w_constr['p'] == "no norm"):
        df = J

    elif (w_constr['name'] == "lasso") or ((w_constr['p'] == "L1") and (w_constr['dir'] == "<=")):
        df = sum(abs(w.iloc[:,0].values) >= 1e-6)

    elif (w_constr['name'] == "simplex") or ((w_constr['p'] == "L1") and (w_constr['dir'] == "==")):
        df = sum(abs(w.iloc[:,0].values) >= 1e-6) - 1

    elif (w_constr['name'] == "ridge") or (w_constr["p"] == "L2"):
        d = numpy.linalg.svd(B)[1]
        d = d[d > 0]
        df = sum(d**2 / (d**2 + w_constr['lambda']))

    # add degrees of freedom coming from C block
    df = df + KM

    return df


def u_sigma_est(u_mean, u_sigma, res, Z, V, index, TT, M, df):
    ZZ = numpy.array(Z)

    if u_sigma == 'HC0':
        vc = 1

    elif u_sigma == 'HC1':
        vc = TT / (TT - df)

    elif u_sigma == 'HC2':
        ZVZinv = numpy.linalg.pinv(ZZ.T.dot(V).dot(ZZ))
        PP = ZZ.dot(ZVZinv).dot(ZZ.T).dot(V)
        vc = 1 / (1 - numpy.diag(PP))

    elif u_sigma == 'HC3':
        ZVZinv = numpy.linalg.pinv(ZZ.T.dot(V).dot(ZZ))
        PP = ZZ.dot(ZVZinv).dot(ZZ.T).dot(V)
        vc = 1 / (1 - numpy.diag(PP))**2

    Omega = numpy.diag(numpy.array((res - u_mean)**2).flatten() * vc)
    Sigma = ZZ.T.dot(V).dot(Omega).dot(V).dot(ZZ) / (TT**2)

    return Sigma, Omega


def cond_pred(y, x, xpreds, method, tau=None):

    if len(x) <= len(x.columns):
        warnings.warn("Consider specifying a less complicated model for e. The number of observations used " +
                      "to parametrically predict moments is smaller than the number of covariates used. Consider " +
                      "reducing either the number of lags (e_lags) or the order of the polynomial (e_order)!")

    if method == 'lm':
        pred = xpreds.dot(sm.OLS(y, x, missing='drop').fit().params)

    elif method == 'qreg':
        qr = sm.QuantReg(endog=y, exog=x)
        pred = numpy.empty(shape=(len(xpreds), 2), dtype='object')
        try:
            pred[:, 0] = qr.fit(q=tau[0]).predict(exog=xpreds)
            pred[:, 1] = qr.fit(q=tau[1]).predict(exog=xpreds)
        except ValueError:
            raise Exception("The design matrix used to model e is not of full-rank! Consider" +
                            "reducing either e_lags or e_order!")
    return pred


def scpi_in_simul(i, beta, Sigma_root, Q, P, J, KM, w_lb_est,
                  w_ub_est, p, p_int, QQ, dire, lb, cores, opt_dict):

    opt_dict = prepareOptions(opt_dict, "scpi")

    zeta = numpy.random.normal(loc=0, scale=1, size=len(beta))
    G = Sigma_root.dot(zeta)

    res_ub = []
    res_lb = []

    x0 = numpy.array([0] * len(beta))

    use_cvxpy = useCVXPY({'Q': QQ, 'dir': dire, 'p': p})

    for hor in range(0, len(P)):
        pt = numpy.array(P.iloc[hor, :])

        if w_lb_est is True:

            if use_cvxpy is True:
                x = cvxpy.Variable(J + KM)
                a = -2 * G - 2 * beta.T.dot(Q)
                d = 2 * G.dot(beta) + beta.dot(Q.dot(beta))

                objective = cvxpy.Minimize(-cvxpy.sum(cvxpy.multiply(pt, x - beta)))

                if lb == 0:
                    constraints = [cvxpy.norm1(x[0:J]) <= QQ,
                                   x[0:J] >= 0,
                                   cvxpy.quad_form(x, Q) + cvxpy.sum(cvxpy.multiply(a, x)) + d <= 0]
                else:
                    constraints = [cvxpy.norm1(x[0:J]) <= QQ,
                                   cvxpy.quad_form(x, Q) + cvxpy.sum(cvxpy.multiply(a, x)) + d <= 0]
                prob = cvxpy.Problem(objective, constraints)
                sol = prob.solve()

                alert = prob.status != 'optimal'

                if alert is True:
                    res_lb.append(numpy.nan)
                else:
                    res_lb.append(sol)

            else:
                opt = nlopt.opt(nlopt.LD_SLSQP, J + KM)
                opt.set_min_objective(lambda x, grad: obj_fun_min(x, grad, pt, beta))

                # add loss function constraint
                opt.add_inequality_constraint(lambda x, grad: single_ineq(x, grad, beta, Q, G), opt_dict['tol_ineq'])

                if dire == "==":  # equality constraint on norm
                    opt.add_equality_constraint(lambda x, grad: norm_equal(x, grad, J, KM, QQ, p_int, dire),
                                                opt_dict['tol_eq'])

                elif dire == "<=":  # inequality constraint on norm
                    opt.add_inequality_constraint(lambda x, grad: norm_equal(x, grad, J, KM, QQ, p_int, dire),
                                                  opt_dict['tol_ineq'])

                opt.set_lower_bounds([lb] * J + [-numpy.inf] * KM)
                opt.set_xtol_rel(opt_dict['xtol_rel'])
                opt.set_xtol_abs(opt_dict['xtol_abs'])
                opt.set_ftol_abs(opt_dict['ftol_abs'])
                opt.set_maxeval(opt_dict['maxeval'])

                x = opt.optimize(x0)

                alert = opt.last_optimize_result() < 0 or opt.last_optimize_result() >= 5
                flag = checkConstraints(x, dire, beta, Q, G, J, KM, QQ, p_int, 1e-2, 1e-2)

                if alert is True or flag is True:
                    res_lb.append(numpy.nan)
                else:
                    res_lb.append(opt.last_optimum_value())

        else:
            res_lb.append(numpy.nan)

        if w_ub_est is True:

            if use_cvxpy is True:
                x = cvxpy.Variable(J + KM)
                a = -2 * G - 2 * beta.T.dot(Q)
                d = 2 * G.dot(beta) + beta.dot(Q.dot(beta))

                objective = cvxpy.Minimize(sum(cvxpy.multiply(pt, x - beta)))

                if lb == 0:
                    constraints = [cvxpy.norm1(x[0:J]) <= QQ,
                                   x[0:J] >= 0,
                                   cvxpy.quad_form(x, Q) + sum(cvxpy.multiply(a, x)) + d <= 0]
                else:
                    constraints = [cvxpy.norm1(x[0:J]) <= QQ,
                                   cvxpy.quad_form(x, Q) + sum(cvxpy.multiply(a, x)) + d <= 0]
                prob = cvxpy.Problem(objective, constraints)
                sol = prob.solve()

                alert = prob.status != 'optimal'

                if alert is True:
                    res_ub.append(numpy.nan)
                else:
                    res_ub.append(-sol)

            else:

                opt = nlopt.opt(nlopt.LD_SLSQP, J + KM)

                opt.set_min_objective(lambda x, grad: obj_fun_max(x, grad, pt, beta))

                # add loss function constraint
                opt.add_inequality_constraint(lambda x, grad: single_ineq(x, grad, beta, Q, G), opt_dict['tol_ineq'])

                if dire == "==":  # equality constraint on norm
                    opt.add_equality_constraint(lambda x, grad: norm_equal(x, grad, J, KM, QQ, p_int, dire),
                                                opt_dict['tol_eq'])

                elif dire == "<=":  # inequality constraint on norm
                    opt.add_inequality_constraint(lambda x, grad: norm_equal(x, grad, J, KM, QQ, p_int, dire),
                                                  opt_dict['tol_ineq'])

                opt.set_lower_bounds([lb] * J + [-numpy.inf] * KM)
                opt.set_xtol_rel(opt_dict['xtol_rel'])
                opt.set_xtol_abs(opt_dict['xtol_abs'])
                opt.set_ftol_abs(opt_dict['ftol_abs'])
                opt.set_maxeval(opt_dict['maxeval'])
                x = opt.optimize(x0)

                alert = opt.last_optimize_result() < 0 or opt.last_optimize_result() >= 5
                flag = checkConstraints(x, dire, beta, Q, G, J, KM, QQ, p_int, 1e-2, 1e-2)

                if alert is True or flag is True:
                    res_ub.append(numpy.nan)
                else:
                    res_ub.append(-opt.last_optimum_value())

        else:
            res_ub.append(numpy.nan)

    res_lb.extend(res_ub)

    return res_lb


def scpi_in(sims, beta, Sigma_root, Q, P, J, KM, w_lb_est,
            w_ub_est, p, p_int, QQ, dire, lb, cores, opt_dict, pass_stata):

    # Progress bar
    iters = round(sims / 10)
    perc = 0

    sims_res = []

    if cores == 1:
        for i in range(sims):
            rem = (i + 1) % iters
            if rem == 0:
                perc = perc + 10
                if pass_stata is False:
                    if any('SPYDER' in name for name in os.environ) and pass_stata is False:
                        print(str(i + 1) + "/" + str(sims) +
                              " iterations completed (" + str(perc) + "%)", end="\n")
                    else:
                        print('\x1b[1A\x1b[2K')
                        print(str(i + 1) + "/" + str(sims) +
                              " iterations completed (" + str(perc) + "%)", end="\r")

            res = scpi_in_simul(i, beta, Sigma_root, Q, P, J, KM, w_lb_est,
                                w_ub_est, p, p_int, QQ, dire, lb, cores, opt_dict)
            sims_res.append(res)

        vsig = numpy.array(sims_res)

    if cores > 1:
        print("Starting parallelization with " + str(cores) + " cores...")
        from dask import compute, delayed, config
        from dask.distributed import Client

        config.set(scheduler='multiprocessing')
        client = Client(n_workers=cores)
        for i in range(sims):
            res = delayed(scpi_in_simul)(i, beta, Sigma_root, Q, P, J, KM, w_lb_est,
                                         w_ub_est, p, p_int, QQ, dire, lb, cores, opt_dict)
            sims_res.append(res)

        vs = compute(sims_res)
        vsig = numpy.array(vs[0])

        client.close()

        shutil.rmtree("dask-worker-space")

        print("")
        print("Closing working clusters...")

    return vsig


def checkConstraints(x, dire, beta, Q, G, J, KM, QQ, p_int, tol_eq, tol_ineq):

    grad = numpy.array([1] * len(x))

    if dire is None:
        flag = single_ineq(x, grad, beta, Q, G) > tol_ineq

    elif dire == "==":
        flag1 = single_ineq(x, grad, beta, Q, G) > tol_ineq
        flag2 = norm_equal(x, grad, J, KM, QQ, p_int, dire) > tol_eq
        flag = flag1 or flag2

    elif dire == "<=":
        flag1 = single_ineq(x, grad, beta, Q, G) > tol_ineq
        flag2 = norm_equal(x, grad, J, KM, QQ, p_int, dire) > tol_ineq
        flag = flag1 or flag2

    return flag


def prepareOptions(opt_dict, prompt):

    opt_names = []
    opt_values = []

    if opt_dict is not None:
        for name, value in opt_dict.items():
            opt_names.append(name)
            opt_values.append(value)
    else:
        opt_dict = {}

    if 'xtol_rel' not in opt_names:
        opt_dict['xtol_rel'] = 1e-8
    if 'xtol_abs' not in opt_names:
        opt_dict['xtol_abs'] = 1e-8
    if 'maxeval' not in opt_names:
        opt_dict['maxeval'] = 5000
    if 'tol_eq' not in opt_names:
        opt_dict['tol_eq'] = 1e-8
    if 'tol_ineq' not in opt_names:
        opt_dict['tol_ineq'] = 1e-8

    if prompt == "scest":
        if 'ftol_rel' not in opt_names:
            opt_dict['ftol_rel'] = 1e-8
        if 'ftol_abs' not in opt_names:
            opt_dict['ftol_abs'] = 1e-8
    else:
        if 'ftol_rel' not in opt_names:
            opt_dict['ftol_rel'] = 1e-4
        if 'ftol_abs' not in opt_names:
            opt_dict['ftol_abs'] = 1e-4

    return opt_dict


def scpi_out(y, x, preds, e_method, alpha, e_lb_est, e_ub_est):
    e_1 = e_2 = None

    if e_lb_est is True or e_ub_est is True:

        if e_method == 'gaussian':
            x_more = pandas.concat([preds, x], axis=0)
            fit = cond_pred(y=y, x=x, xpreds=x_more, method='lm')
            e_mean = fit[:len(preds)]
            y_fit = fit[len(preds):]
            y_var = numpy.log((y - y_fit)**2)
            var_pred = cond_pred(y=y_var, x=x, xpreds=x_more, method='lm')
            e_sig2 = numpy.exp(var_pred[:len(preds)])

            eps = numpy.sqrt(-numpy.log(alpha) * 2 * e_sig2)
            lb = e_mean - eps
            ub = e_mean + eps

            e_1 = e_mean
            e_2 = e_sig2

            lb = lb.to_numpy().reshape(len(lb), 1)
            ub = ub.to_numpy().reshape(len(ub), 1)

        elif e_method == 'ls':
            x_more = pandas.concat([preds, x], axis=0)
            fit = cond_pred(y=y, x=x, xpreds=x_more, method='lm')
            e_mean = fit[:len(preds)]
            y_fit = fit[len(preds):]
            y_var = numpy.log((y - y_fit)**2)
            var_pred = cond_pred(y=y_var, x=x, xpreds=x_more, method='lm')
            e_sig = numpy.sqrt(numpy.exp(var_pred[:len(preds)]))
            y_st = (y - y_fit) / numpy.sqrt(numpy.exp(var_pred[len(preds):]))
            lb = e_mean + e_sig * y_st.quantile(q=alpha)
            ub = e_mean + e_sig * y_st.quantile(q=(1 - alpha))

            lb = lb.to_numpy().reshape(len(lb), 1)
            ub = ub.to_numpy().reshape(len(ub), 1)

            e_1 = e_mean
            e_2 = e_sig**2

        elif e_method == 'qreg':
            e_pred = cond_pred(y=y, x=x, xpreds=preds,
                               method='qreg', tau=[alpha, 1 - alpha])
            lb = e_pred[:, [0]]
            ub = e_pred[:, [1]]

    else:
        lb = None
        ub = None

    return lb, ub, e_1, e_2


def regularize_w(rho, rho_max, res, B, C, coig_data, T0_tot):
    if rho == 'type-1':
        ssr = (res - res.mean())**2
        sigma_u = sqrt(ssr.mean())
        sigma_bj = min(B.std(axis=0))
        CC = sigma_u / sigma_bj
    elif rho == 'type-2':
        ssr = (res - res.mean())**2
        sigma_u = sqrt(ssr.mean())
        sigma_bj2 = min(B.var(axis=0))
        sigma_bj = max(B.var(axis=0))
        CC = sigma_u * sigma_bj / sigma_bj2
    elif rho == 'type-3':
        sigma_bj2 = min(B.var(axis=0))
        tempdf = pandas.concat([res, B], axis=1)
        sigma_bju = max(tempdf.cov().iloc[:, 0])
        CC = sigma_bju / sigma_bj2

    if coig_data is True:
        c = 1
    else:
        c = 0.5

    rho = (CC * (log(T0_tot))**c) / (sqrt(T0_tot))

    if rho_max is not None:
        rho = min(rho, rho_max)

    return rho


def regularize_check(w, index_w, rho):
    # If regularization is ill-behaved just select the first weight
    if len(index_w) == 0:
        sel = w.rank(ascending=False) <= 1
        index_w = w.loc[sel[0], ].index.get_level_values(0).tolist()
        print("Warning: regularization paramater was too high (" + str(round(rho, 3)) + "). "
              + "We set it so that at least one component in w is non-zero.")
    return(index_w)


def local_geom(w_constr, rho, rho_max, res, B, C, coig_data, T0_tot, J, w):
    Q = w_constr['Q']
    if isinstance(rho, str):
        rho = regularize_w(rho, rho_max, res, B, C, coig_data, T0_tot)

    if (w_constr['name'] == "simplex") | ((w_constr['p'] == "L1") & (w_constr['dir'] == "==")):
        index_w = w.loc[w[0] > rho, ].index.get_level_values(0).tolist()
        index_w = regularize_check(w, index_w, rho)
        w_star = deepcopy(w)
        to_regularize = [col for col in B.columns if col not in index_w]
        w_star.loc[to_regularize, ] = 0

        Q_star = float(numpy.sum(w_star))

    elif (w_constr['name'] == "lasso") | ((w_constr['p'] == "L1") & (w_constr['dir'] == "<=")):
        l1 = float(numpy.sum(abs(w)))
        if (l1 >= Q - rho * sqrt(J)) & (l1 <= Q):
            Q_star = l1
        else:
            Q_star = Q

        index_w = w.loc[abs(w[0]) > rho, ].index.get_level_values(0).tolist()
        index_w = regularize_check(w, index_w, rho)
        w_star = deepcopy(w)
        to_regularize = [col for col in B.columns if col not in index_w]
        w_star.loc[to_regularize, ] = 0

    elif (w_constr['name'] == "ridge") or (w_constr['p'] == "L2"):
        l2 = float(sqrt(numpy.sum(w**2)))
        if (l2 >= Q - rho) & (l2 <= Q):
            Q_star = l2
        else:
            Q_star = Q

        w_star = deepcopy(w)
        index_w = B.columns

    else:
        Q_star = Q
        w_star = w
        index_w = B.columns

    w_constr['Q'] = Q_star

    return w_constr, w_star, index_w, rho, Q_star

def executionTime(T0, T1, J, cores, sims, name):
    Ttot = sum(T0.values())
    tincr = Ttot / 1000

    coefsJ = numpy.array([-0.54755616, 0.09985644])

    time = numpy.array([1, J]) @ coefsJ
    time = time * sims / 10
    time = time / cores
    time = time * tincr
    time = time * T1
    time = time / 60
    time = 2 * ceil(time)

    if name == "lasso":
        time = 3 * time

    if time < 1:
        toprint = "Maximum expected execution time: less than a minute."
    elif time == 1:
        toprint = "Maximum expected execution time: " + str(time) + " minute."
    else:
        toprint = "Maximum expected execution time: " + str(time) + " minutes."

    print(toprint)

def useCVXPY(w_constr):
    flag = False
    if w_constr['p'] == "L1" and w_constr['dir'] == "<=":
        flag = True

    return flag



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 16 11:08:57 2021

@author: Filippo Palomba
"""
# Temporary code to suppress pandas FutureWarning
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import pandas
import numpy
from .funs import w_constr_prep, b_est
from .scplot import scplot


def scest(df, w_constr=None, V=None, opt_dict=None, plot=False):

    '''

    Parameters
    ----------
    df : scdata_output
        a class scdata_output object, obtained by calling scdata

    w_constr : dictionary
        a dictionary specifying the constraint set the estimated weights of the donors must belong to.
        w_constr can contain up to four objects:
        - p, a string indicating the norm to be used (p should be one of "no norm", "L1", and "L2")
        - dir, a string indicating whether the constraint on the norm is an equality ("==") or inequality ("<=")
        - Q, a scalar defining the value of the constraint on the norm
        - lb, a scalar defining the lower bound on the weights. It can be either 0 or -numpy.inf.
        - name, a character selecting one of the default proposals.

    V : numpy.array, default numpy.identity
        a weighting matrix to be used when minimizing the sum of squared residuals.
        The default is the identity matrix, so equal weight is given to all observations.

    opt_dict : dictionary
        a dictionary specifying the stopping criteria used by the underling optimizer (nlopt) for point estimation.
        The default is a sequential quadratic programming (SQP) algorithm for nonlinearly constrained gradient-based
        optimization ('SLSQP'). In case a lasso-type constraint is implemented, cvxpy is used for optimization.
        More information on the stopping criteria can be obtained reading the official documentation at
        https://www.cvxpy.org/. The default values are 'maxeval = 5000', 'xtol_rel = 1e-8', 'xtol_abs = 1e-8',
        'ftol_rel = 1e-8', 'ftol_abs = 1e-8', 'tol_eq = 1e-8', and 'tol_ineq = 1e-8'.

    plot : bool, default False
        a logical specifying whether scplot should be called and a plot saved in the current working directory. For more
        options see scplot.

    Returns
    -------
    The function returns an object of class `scest_output' containing the following objects

    w : pandas.DataFrame
        a dataframe containing the weights of the donors.

    r : pandas.DataFrame
        a dataframe containing the values of the covariates used for adjustment.

    b : pandas.DataFrame
        a dataframe containing w and r.

    Y_pre_fit : pandas.DataFrame
        a dataframe containing the estimated pre-treatment outcome for the SC unit.

    Y_post_fit : pandas.DataFrame
        a dataframe containing the estimated post-treatment outcome for the SC unit.

    A_hat : pandas.DataFrame
        a dataframe containing the predicted values of the features of the treated unit.

    res : pandas.DataFrame
        a dataframe containing the residuals A - A_hat.

    V : numpy.array
        an array containing the weighting matrix used in estimation.

    w_constr : dictionary
        a dictionary containing the specifics of the constraint set used on the weights.

    A : pandas.DataFrame
        a dataframe containing pre-treatment features of the treated unit.

    B : pandas.DataFrame
        a dataframe containing pre-treatment features of the control units.

    C : pandas.DataFrame
        a dataframe containing covariates for adjustment.

    P : pandas.DataFrame
        a dataframe whose rows are the vectors used to predict the out-of-sample series for the synthetic unit.

    Y_pre : pandas.DataFrame
        a dataframe containing the pre-treatment outcome of the treated unit.

    Y_post : pandas.DataFrame
        a dataframe containing the post-treatment outcome of the treated unit.

    Y_donors : pandas.DataFrame
        a dataframe containing the pre-treatment outcome of the control units.

    J : int
        the number of control units

    K : array
        a numeric array with the number of covariates used for adjustment for each feature

    KM : int
        the total number of covariates used for adjustment

    M : int
        number of features

    period_pre : array
        a numeric array with the pre-treatment period

    period_post : array
        a numeric array with the post-treatment period

    T0_features : array
        a numeric array with the number of periods used in estimation for each feature

    T1_outcome : int
        the number of post-treatment periods

    glob_cons : bool
        for internal use only

    out_in_features : bool
        for internal use only

    cointegrated_data: bool
        logical indicating whether the data has been treated as cointegrated.

    References
    ----------
    Abadie, A. (2021), “Using Synthetic Controls: Feasibility, Data Requirements, and Methodological
    Aspects,” Journal of Economic Literature, 59, 391-425.

    Cattaneo, M. D., Feng, Y., and Titiunik, R. (2021), “Prediction Intervals for Synthetic Control
    Methods,” Journal of the American Statistical Association, 116, 1865-1880.

    Cattaneo, M. D., Palomba, F., Feng, Y., and Titiunik, R. (2022), “scpi: Uncertainty Quantification for
    Synthetic Control Estimators”.

    See Also
    --------
    scdata, scpi, scplot

    '''

    if df.__class__.__name__ != 'scdata_output':
        raise Exception("df should be the object returned by running scdata!")

    w_names = []
    w_values = []
    if w_constr is not None:
        if not isinstance(w_constr, dict):
            raise Exception("w_constr should be an object of type dict!")

        for name, value in w_constr.items():
            w_names.append(name)
            w_values.append(value)

        if 'name' in w_names:
            if not w_constr['name'] in ["simplex", "lasso", "ridge", "ols"]:
                raise Exception("If 'name' is specified in w_constr, then it should be chosen" +
                                " among 'simplex', 'lasso', 'ridge', 'ols'.")
        else:
            w_constr['name'] = None

    if opt_dict is not None:
        if not isinstance(opt_dict, dict):
            raise Exception("opt_dict should be an object of type dict!")

    # Data matrices and specs
    A = df.A
    B = df.B
    C = df.C
    P = df.P
    Z = pandas.concat([B, C], axis=1)
    Y_donors = df.Y_donors

    J = df.J
    K = df.K
    KM = df.KM
    M = df.M
    T0_features = df.T0_features
    out_in_features = df.out_in_features
    outcome_var = df.outcome_var

    ##########################
    # Set up the estimation problem

    # Create weighting matrix
    if V is not None:
        if not isinstance(V, (pandas.DataFrame, numpy.ndarray)):
            raise Exception("The object V should be a dataframe or a matrix!")

        V_shape = numpy.shape(V)
        if V_shape[0] != len(B) or V_shape[1] != len(B):
            raise Exception("The object V should be of shape (" + str(len(B)) +
                            "," + str(len(B)) + ")!")

    if V is None:
        V = numpy.identity(len(A))

    # Set constraint on weights
    w_constr = w_constr_prep(w_constr, w_names, A, Z, V, J, KM)

    # if not isinstance(w_constr['lb'], (int, float)):
    #    if not len(w_constr['lb']) != J:
    #            raise Exception("The length of 'lb' should be either one or the number of donors.")

    # Estimate SC
    result = b_est(A=A, Z=Z, J=J, KM=KM, w_constr=w_constr, V=V, opt_dict=opt_dict)
    b = pandas.DataFrame(result, index=Z.columns)

    ##########################
    # Create useful objects

    # Store point estimates
    if KM == 0:
        w = b
        r = pandas.DataFrame(None)
    else:
        w = b[0:len(B.columns)]
        r = b[len(B.columns):]

    # Fitted values and residuals
    A_hat = Z.dot(b)
    A_hat.columns = A.columns
    res = A - A_hat

    # Pre-treatment fit of outcome of interest
    if out_in_features is True:   # outcome is among features
        fit_pre = A_hat.iloc[:T0_features[outcome_var], ].reset_index(level='feature', drop=True)

    else:  # outcome is not among features
        fit_pre = Y_donors.dot(w)

    fit_post = P.dot(b)

    fit_pre.columns = A.columns
    fit_post.columns = A.columns
    A_hat.columns = A.columns

    ##################################################
    # Plot

    if plot is True:
        to_plot = scest_output(b=b, w=w, r=r, Y_pre_fit=fit_pre,
                               Y_post_fit=fit_post, A_hat=A_hat, res=res,
                               V=V, w_constr=w_constr, A=A, B=B, C=C,
                               P=P, Y_pre=df.Y_pre, Y_post=df.Y_post,
                               Y_donors=Y_donors, J=J, K=K, KM=KM,
                               M=M,
                               cointegrated_data=df.cointegrated_data,
                               period_pre=df.period_pre,
                               period_post=df.period_post,
                               T0_features=T0_features,
                               T1_outcome=df.T1_outcome,
                               outcome_var=df.outcome_var,
                               features=df.features,
                               glob_cons=df.glob_cons,
                               out_in_features=out_in_features,
                               plotres=None)

        plotres = scplot(result=to_plot)

    else:
        plotres = None

    return scest_output(b=b,
                        w=w,
                        r=r,
                        Y_pre_fit=fit_pre,
                        Y_post_fit=fit_post,
                        A_hat=A_hat,
                        res=res,
                        V=V,
                        w_constr=w_constr,
                        A=A,
                        B=B,
                        C=C,
                        P=P,
                        Y_pre=df.Y_pre,
                        Y_post=df.Y_post,
                        Y_donors=Y_donors,
                        J=J,
                        K=K,
                        KM=KM,
                        M=M,
                        cointegrated_data=df.cointegrated_data,
                        period_pre=df.period_pre,
                        period_post=df.period_post,
                        T0_features=T0_features,
                        T1_outcome=df.T1_outcome,
                        outcome_var=df.outcome_var,
                        features=df.features,
                        glob_cons=df.glob_cons,
                        out_in_features=out_in_features,
                        plotres=plotres)

class scest_output:
    def __init__(self, b, w, r, Y_pre_fit, Y_post_fit, A_hat, res, V, w_constr,
                 A, B, C, P, Y_pre, Y_post, Y_donors, J, K, KM, M,
                 cointegrated_data, period_pre, period_post, T0_features,
                 T1_outcome, features, outcome_var, glob_cons, out_in_features, plotres):

        self.b = b
        self.w = w
        self.r = r
        self.Y_pre_fit = Y_pre_fit
        self.Y_post_fit = Y_post_fit
        self.A_hat = A_hat
        self.res = res
        self.V = V
        self.w_constr = w_constr
        self.A = A
        self.B = B
        self.C = C
        self.P = P
        self.Y_pre = Y_pre
        self.Y_post = Y_post
        self.Y_donors = Y_donors
        self.J = J
        self.K = K
        self.KM = KM
        self.M = M
        self.cointegrated_data = cointegrated_data
        self.period_pre = period_pre
        self.period_post = period_post
        self.T0_features = T0_features
        self.T1_outcome = T1_outcome
        self.features = features
        self.outcome_var = outcome_var
        self.glob_cons = glob_cons
        self.out_in_features = out_in_features
        self.plotres = plotres

    def __repr__(self):

        # Prepare objects to print
        fw = 30
        if self.M == 1:
            fw_r = 35
        if self.M > 1:
            fw_r = 40

        constr = self.w_constr['name']
        if self.w_constr['Q'] is not None:
            Qsize = round(self.w_constr['Q'], 3)
        else:
            Qsize = "-"
        tr_unit = self.A.columns.values.tolist()
        pt_in = self.period_pre[0]
        pt_fi = self.period_pre[len(self.period_pre) - 1]
        ppre = str(pt_in) + '-' + str(pt_fi)

        Weights = self.w.rename(columns={0: 'Weights'}, inplace=False)
        Weights = round(Weights, 3)
        activew = len(Weights.loc[abs(Weights['Weights']) > 0])

        if self.KM > 0:
            Covariates = self.r.rename(columns={0: 'Covariates'}, inplace=False)
            Covariates = round(Covariates, 3)

        if constr is None:
            constr = "User Provided"

        # Print stuff
        print('-----------------------------------------------------------------------')
        print('Call: scest')
        print('Synthetic Control Estimation - Setup')
        print('')

        print('Constraint Type:'.ljust(fw), str(constr).rjust(fw_r))
        print('Constraint Size (Q):'.ljust(fw), str(Qsize).rjust(fw_r))
        print('Treated Unit:'.ljust(fw), str(tr_unit[0]).rjust(fw_r))
        print('Size of the donor pool:'.ljust(fw), str(self.J).rjust(fw_r))
        print('Features'.ljust(fw), str(self.M).rjust(fw_r))
        print('Pre-treatment period'.ljust(fw), str(ppre).rjust(fw_r))

        if self.M == 1:
            T0 = [c for c in self.T0_features.values()]

            print('Pre-treatment periods used in estimation:'.ljust(fw),
                  str(T0[0]).rjust(fw_r - 11))
            print('Covariates used for adjustment:'.ljust(fw),
                  str(self.KM).rjust(fw_r - 1))
        else:
            toprint = pandas.DataFrame.from_dict(self.T0_features, orient='index').reset_index()
            toprint.columns = ['Feature', 'Observations']
            print('Pre-treatment periods used in estimation per feature:'.ljust(fw))
            print(toprint.to_string(index=False))

            toprint = pandas.DataFrame(self.K)
            print('Covariates used for adjustment per feature:'.ljust(fw))
            print(toprint.to_string(index=False))

        print('')
        print('Synthetic Control Estimation - Results')
        print('')
        print('Active donors:', activew)
        print('')
        print('Coefficients:')
        print(Weights)
        if self.KM > 0:
            print('')
            print(Covariates)

        return ''



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 17 14:46:14 2021

@author: Filippo Palomba
"""
# Temporary code to suppress pandas FutureWarning
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import pandas
import numpy
import warnings
from copy import deepcopy
import multiprocessing as mp
from sklearn.linear_model import LinearRegression
from .funs import local_geom, u_des_prep, e_des_prep, complete_cases, df_EST, u_sigma_est, scpi_in, scpi_out
from .funs import executionTime, createPoly, DUflexGet
from .scest import scest
from .scplot import scplot
from scipy.linalg import sqrtm

def scpi(data,
         w_constr=None,
         V=None,
         P=None,
         u_missp=True,
         u_sigma="HC1",
         u_order=1,
         u_lags=0,
         u_design=None,
         u_alpha=0.05,
         e_method="all",
         e_order=1,
         e_lags=0,
         e_design=None,
         e_alpha=0.05,
         sims=200,
         rho=None,
         rho_max=None,
         cores=None,
         plot=False,
         w_bounds=None,
         e_bounds=None,
         opt_dict_est=None,
         opt_dict_inf=None,
         pass_stata=False):

    '''
    Parameters
    ----------
    data : scdata_output
        a class scdata_output object, obtained by calling scdata

    w_constr : dictionary
        a dictionary specifying the constraint set the estimated weights of the donors must belong to.
        w_constr can contain up to five objects:
        1. p, a scalar indicating the norm to be used (p should be one of "no norm", "L1", and "L2")
        2. dir, a string indicating whether the constraint on the norm is an equality ("==") or inequality ("<=")
        3. Q, a scalar defining the value of the constraint on the norm
        4. lb, a scalar defining the lower bound on the weights. It can be either 0 or -numpy.inf.
        5. name, a character selecting one of the default proposals.

    V : numpy.array, default numpy.identity
        an array specifying the weighting matrix to be used when minimizing the sum of squared residuals.
        The default is the identity matrix, so equal weight is given to all observations.

    P : numpy.array, default None
        a T_1 x (J+K_1) array containing the design matrix to be used to obtain the predicted.
        post-intervention outcome of the synthetic control unit. T_1 is the number of post-treatment periods,
        J is the size of the donor pool, and K_1 is the number of covariates used for adjustment in
        the outcome equation.

    u_missp : bool, default True
        a logical indicating if misspecification should be taken into account when dealing with u.

    u_sigma : str, default "HC1"
        a string specifying the type of variance-covariance estimator to be used when estimating
        the conditional variance of u. Available choices are HC0, HC1, HC2, and HC3.

    u_order : int, default 1
        an integer that sets the order of the polynomial in B when predicting moments of u.

    u_lags : int, default 0
        an integer that sets the number of lags of B when predicting moments of u.

    u_design : numpy.array, default None
        an array with the same number of rows of A and B and whose columns specify the design matrix
        to be used when modeling the estimated pseudo-true residuals u.

    u_alpha : float, default 0.05
        the confidence level for in-sample uncertainty.

    e_method : str, default "all"
        a string selecting the method to be used in quantifying out-of-sample uncertainty among:
        "gaussian" which uses conditional subgaussian bounds; "ls" which specifies a location-scale model for u; "qreg"
        which employs a quantile regressions to get the conditional bounds; "all" uses each one of the previous methods.

    e_order : int, default 1
        an integer that sets the order of the polynomial in B when predicting moments of e.

    e_lags: int, default 0
        a scalar that sets the number of lags of B when predicting moments of e.

    e_design : numpy.array, default None
        an array with the same number of rows of A and B and whose columns specify the design matrix
        to be used when modeling the estimated out-of-sample residuals e.

    e_alpha : float, default 0.05
        an integer specifying the confidence level for out-of-sample uncertainty.

    sims : int, default 200
        an integer providing the number of simulations to be used in quantifying in-sample uncertainty.

    rho : float/str, default 'type-1'
        a float specifying the regularizing parameter that imposes sparsity on the estimated vector of weights. If
        rho = 'type-1', then the tuning parameter is computed based on optimization inequalities. Other options are
        'type-2', and 'type-3'. See the software article for more information.

    rho_max : float, default 1
        a float indicating the maximum value attainable by the tuning parameter rho.

    cores : integer, default multiprocessing.cpu_count() - 1
        number of cores to be used by the command. The default is half the cores available.

    plot : bool, default False
        a logical specifying whether scplot should be called and a plot saved in the current working directory.
        For more options see scplot.

    w_bounds : numpy.array
        a T1 x 2 array with the user-provided bounds on beta. If w_bounds is provided, then
        the quantification of in-sample uncertainty is skipped. It is possible to provide only the lower bound or the
        upper bound by filling the other column with NAs.

    e_bounds : numpy.array
        a T1 x 2 array with the user-provided bounds on e. If e_bounds is provided, then
        the quantification of out-of-sample uncertainty is skipped. It is possible to provide only the lower bound or
        the upper bound by filling the other column with NAs.

    opt_dict_est : dictionary
        a dictionary specifying the stopping criteria used by the underling optimizer (nlopt) for point estimation.
        The default is a sequential quadratic programming (SQP) algorithm for nonlinearly constrained gradient-based
        optimization ('SLSQP'). In case a lasso-type constraint is implemented, cvxpy is used for optimization.
        More information on the stopping criteria can be obtained reading the official documentation at
        https://www.cvxpy.org/. The default values are 'maxeval = 5000', 'xtol_rel = 1e-8', 'xtol_abs = 1e-8',
        'ftol_rel = 1e-4', 'ftol_abs = 1e-4', 'tol_eq = 1e-8', and 'tol_ineq = 1e-8'.

    opt_dict_inf : dictionary
        same as above but for inference purposes. The default values are 'maxeval = 5000', 'xtol_rel = 1e-8',
        'xtol_abs = 1e-8', 'ftol_rel = 1e-4', 'ftol_abs = 1e-4', 'tol_eq = 1e-8', and 'tol_ineq = 1e-8'.

    pass_stat : bool
        for internal use only.

    Returns
    -------
    The function returns an object of class `scpi_output' containing the following objects

    w : pandas.DataFrame
        a dataframe containing the weights of the donors.

    r : pandas.DataFrame
        a dataframe containing the values of the covariates used for adjustment.

    b : pandas.DataFrame
        a dataframe containing w and r.

    Y_pre_fit : pandas.DataFrame
        a dataframe containing the estimated pre-treatment outcome for the SC unit.

    Y_post_fit : pandas.DataFrame
        a dataframe containing the estimated post-treatment outcome for the SC unit.

    A_hat : pandas.DataFrame
        a dataframe containing the predicted values of the features of the treated unit.

    res : pandas.DataFrame
        a dataframe containing the residuals A - A_hat.

    V : numpy.array
        an array containing the weighting matrix used in estimation.

    w_constr : dictionary
        a dictionary containing the specifics of the constraint set used on the weights.

    A : pandas.DataFrame
        a dataframe containing pre-treatment features of the treated unit.

    B : pandas.DataFrame
        a dataframe containing pre-treatment features of the control units.

    C : pandas.DataFrame
        a dataframe containing covariates for adjustment.

    P : pandas.DataFrame
        a dataframe whose rows are the vectors used to predict the out-of-sample series for the synthetic unit.

    Y_pre : pandas.DataFrame
        a dataframe containing the pre-treatment outcome of the treated unit.

    Y_post : pandas.DataFrame
        a dataframe containing the post-treatment outcome of the treated unit.

    Y_donors : pandas.DataFrame
        a dataframe containing the pre-treatment outcome of the control units.

    J : int
        the number of control units

    K : array
        a numeric array with the number of covariates used for adjustment for each feature

    KM : int
        the total number of covariates used for adjustment

    M : int
        number of features

    period_pre : array
        a numeric array with the pre-treatment period

    period_post : array
        a numeric array with the post-treatment period

    T0_features : array
        a numeric array with the number of periods used in estimation for each feature

    T1_outcome : int
        the number of post-treatment periods

    glob_cons : bool
        for internal use only

    out_in_features : bool
        for internal use only

    cointegrated_data: bool
        logical indicating whether the data has been treated as cointegrated.

    CI_in_sample : pandas.DataFrame
        a dataframe containing the prediction intervals taking only in-sample uncertainty in to account.

    CI_all_gaussian : pandas.DataFrame
        a dataframe containing the prediction intervals taking out-of-sample uncertainty in to account.

    CI_all_ls : pandas.DataFrame
        a dataframe containing the prediction intervals taking out-of-sample uncertainty in to account.

    CI_all_qreg : pandas.DataFrame
        a dataframe containing the prediction intervals taking out-of-sample uncertainty in to account.

    Sigma : numpy.array
        an array containing the estimated variance-covariance Sigma.

    u_mean : numpy.array
        an array containing the estimated conditional mean of the the pseudo-residuals u.

    u_var : numpy.array
        an array containing the estimated conditional variance-covariance of the pseudo-residuals u.

    e_mean : numpy.array
        an array containing the estimated conditional mean of the out-of-sample error e.

    e_var : numpy.array
        an array containing the estimated conditional variance of the out-of-sample error e.

    u_missp : bool
        a logical indicating whether the model has been treated as misspecified or not.

    u_lags : int
        an integer containing the number of lags in B used in predicting moments of the pseudo-residuals u.

    u_order : int
        an integer containing the order of the polynomial in B used in predicting moments of the pseudo-residuals u.

    u_sigma : str
        a string indicating the estimator used for Sigma.

    u_user : bool
        a logical indicating whether the design matrix to predict moments of u was user-provided.

    u_alpha : float
        a float indicating the confidence level used for in-sample uncertainty.

    e_method : str
        a string indicating the specification used to predict moments of the out-of-sample error e.

    e_lags : int
        an integer containing the number of lags in B used in predicting moments of the pseudo-residuals u.

    e_order : int
        an integer containing the number of lags in B used in predicting moments of the pseudo-residuals u.

    e_user : bool
        a logical indicating whether the design matrix to predict moments of e was user-provided.

    e_alpha : float
        a float indicating the confidence level used for out-of-sample uncertainty.

    rho : str/float
        an integer specifying the estimated regularizing parameter that imposes sparsity on
        the estimated vector of weights.

    sims : int
        an integer indicating the number of simulations.

    failed_sims : numpy.array
        an array containing the number of failed simulations per post-treatment period to estimate
        lower and upper bounds.


    References
    ----------
    Abadie, A. (2021), “Using Synthetic Controls: Feasibility, Data Requirements, and Methodological
    Aspects,” Journal of Economic Literature, 59, 391-425.

    Cattaneo, M. D., Feng, Y., and Titiunik, R. (2021), “Prediction Intervals for Synthetic Control
    Methods,” Journal of the American Statistical Association, 116, 1865-1880.

    Cattaneo, M. D., Palomba, F., Feng, Y., and Titiunik, R. (2022), “scpi: Uncertainty Quantification for
    Synthetic Control Estimators”.

    See Also
    --------
    scdata, scest, splot

    '''

    if data.__class__.__name__ != 'scdata_output':
        raise Exception("data should be the object returned by running scdata!")

    ######################################
    # Estimation of synthetic weights
    if pass_stata is False:
        print("-----------------------------------------------")
        print("Estimating Weights...")
    sc_pred = scest(df=data, w_constr=w_constr, V=V, opt_dict=opt_dict_est)

    ######################################
    # Retrieve processed data from scest

    A = sc_pred.A                           # Features of treated unit
    B = sc_pred.B                           # Features of control units
    C = sc_pred.C                           # Covariates for adjustment
    Z = pandas.concat([B, C], axis=1)       # B and C column-bind
    Y_donors = sc_pred.Y_donors             # Outcome variable of control units
    K = sc_pred.K                           # Number of covs for adjustment per feature
    KM = sc_pred.KM                         # Dimension of r (total number of covs for adj)
    J = sc_pred.J                           # Number of donors
    M = sc_pred.M                           # Number of features
    T0 = sc_pred.T0_features                # Time periods used per feature
    T1 = sc_pred.T1_outcome                 # Number of out-of-sample periods
    outcome_var = sc_pred.outcome_var       # Name of outcome variable
    constant = sc_pred.glob_cons            # Logical indicating whether a constant is included
    out_feat = sc_pred.out_in_features      # Logical indicating whether the outcome variable is among features
    coig_data = sc_pred.cointegrated_data   # Logical indicating whether B is cointegrated
    w_constr = deepcopy(sc_pred.w_constr)   # Constraints on w
    V = sc_pred.V                           # Weighting matrix
    w = deepcopy(sc_pred.w)                 # Estimated vector of weights
    r = sc_pred.r                           # Estimated coefficients of covariates
    Y_post_fit = sc_pred.Y_post_fit         # Estimated post-treatment outcome for SC unit
    res = sc_pred.res                       # Residuals from estimation
    T0_tot = sum(T0.values())               # Total number of observations used in estimation

    if P is None:           # Matrix for out-of-sample prediction
        P = sc_pred.P
    else:                   # User-provided prediction matrix P (should be T1 by (J+KM))
        if not isinstance(P, (pandas.DataFrame, numpy.ndarray)):
            raise Exception("The object P should be a dataframe or an array!")

        if isinstance(P, numpy.ndarray):
            P = pandas.DataFrame(P,
                                 index=Y_post_fit.index,
                                 columns=Y_post_fit.columns)

        P_shape = numpy.shape(P)
        if P_shape[0] != T1:
            raise Exception("The object P has " + str(P_shape[0]) + " rows, when" +
                            " instead " + str(T1) + " were expected (i.e. the number of post-intervention periods)!")

        if P_shape[1] != (J + int(K.iloc[0, 0])):
            raise Exception("The object P has " + str(P_shape[1]) + " columns, when" +
                            " instead " + str(J + int(K.iloc[0, 0])) + " were expected (i.e. the size of the donor " +
                            "pool plus the number of covariates used in adjustment in the outcome equation)!")

        # Add zeros to avoid loading the coefficient of covs used for adj in other eqs
        if (KM - int(K.iloc[0, 0])) > 0:
            zeros = numpy.zeros((T1, sum(K.iloc[1:, ])))
            P = pandas.DataFrame(numpy.c_[P, zeros])

    if not isinstance(u_sigma, str):
        raise Exception("The object u_sigma should be of type character!")
    else:
        if u_sigma not in ['HC0', 'HC1', 'HC2', 'HC3']:
            raise Exception("Supported variance estimators are 'HC0','HC1','HC2','HC3'.")

    if w_bounds is not None:
        if not isinstance(w_bounds, (pandas.DataFrame, numpy.ndarray)):
            raise Exception("The object w_bounds should be a dataframe or an array!")

        if isinstance(w_bounds, numpy.ndarray):
            w_bounds = pandas.DataFrame(w_bounds,
                                        index=Y_post_fit.index,
                                        columns=Y_post_fit.columns)

        w_bounds_shape = numpy.shape(w_bounds)
        if w_bounds_shape[0] != len(Y_post_fit):
            raise Exception("w.bounds should be a matrix with two columns: the first column for the " +
                            "lower bound, the second for the upper. In case you don't want to specify " +
                            "the lower or the upper bound just fill the specific column with NAs.")

        if w_bounds_shape[1] != len(Y_post_fit):
            raise Exception("w.bounds should be a matrix with " + str(len(Y_post_fit))
                            + " rows (i.e. the number of post-intervention periods).")

    if sims < 10:
        raise Exception("The number of simulations needs to be larger or equal than 10!")

    if e_bounds is not None:
        if not isinstance(e_bounds, (pandas.DataFrame, numpy.ndarray)):
            raise Exception("The object e_bounds should be a dataframe or an array!")

        if isinstance(e_bounds, numpy.ndarray):
            e_bounds = pandas.DataFrame(e_bounds,
                                        index=Y_post_fit.index,
                                        columns=Y_post_fit.columns)

        e_bounds_shape = numpy.shape(e_bounds)
        if e_bounds_shape[0] != len(Y_post_fit):
            raise Exception("w.e_bounds should be a matrix with two columns: the first column for the " +
                            "lower bound, the second for the upper. In case you don't want to specify " +
                            "the lower or the upper bound just fill the specific column with NAs.")

        if e_bounds_shape[1] != len(Y_post_fit):
            raise Exception("e_bounds should be a matrix with " + str(len(Y_post_fit))
                            + " rows (i.e. the number of post-intervention periods).")

    # Check rho
    if rho is not None:
        if isinstance(rho, str):
            if rho not in ['type-1', 'type-2', 'type-3']:
                raise Exception("When not a scalar, 'rho' must be 'type-1', 'type-2', or 'type-3'.")
    else:
        rho = 'type-1'

    # Check on number of cores
    if cores is not None:
        n_cores = mp.cpu_count()
        if cores > n_cores:
            raise Exception("You selected " + str(cores) + " cores, but only " + str(n_cores) +
                            " are available on your machine!")
    else:
        cores = mp.cpu_count() - 1

    if pass_stata is False:
        print("Quantifying Uncertainty")
        executionTime(T0, T1, J, cores, sims, w_constr['name'])
        print(" ")

    ######################################
    ######################################
    # Estimate In-Sample Uncertainty
    ######################################
    ######################################

    # Regularize w
    w_constr_inf, w_star, index_w, rho, Q_star = local_geom(w_constr, rho,
                                                            rho_max, res, B, C, coig_data, T0_tot, J, w)

    # Create an index that selects all non-zero weights and additional covariates
    if isinstance(index_w, pandas.core.indexes.base.Index):
        index = deepcopy(index_w.tolist())
    else:
        index = deepcopy(index_w)
    index.extend(C.columns.tolist())
    beta = pandas.concat([w_star, r], axis=0)

    ##########################################################################
    # Prepare design matrix for in-sample uncertainty
    u_des_0 = u_des_prep(B, C, Z, u_order, u_lags, coig_data, T0_tot, M, constant, index,
                         index_w, u_design, res)

    ##########################################################################
    # Prepare design matrix for out_of-sample uncertainty
    e_res, e_des_0, e_des_1 = e_des_prep(B, C, Z, P, e_order, e_lags, res, sc_pred, Y_donors, out_feat, J, M, index,
                                         index_w, coig_data, T0, T0_tot, T1, outcome_var, constant, e_design)

    ###########################################################################
    # Remove NA - In sample uncertainty
    X = pandas.concat([A, res, u_des_0, Z], axis=1)
    tosel = complete_cases(X)
    X_na = X.loc[tosel, ]

    j2 = 1
    j3 = j2 + 1
    j4 = j3 + len(u_des_0.columns)

    res_na = X_na.iloc[:, j2:j3]
    u_des_0_na = X_na.iloc[:, j3:j4]
    Z_na = X_na.iloc[:, j4:]

    V_na = V[numpy.ix_(tosel, tosel)]

    # Effective number of observation used for inference (not yet adjusted for df used)
    TT = len(Z_na)

    # Remove NA - In sample uncertainty
    X = pandas.concat([e_res, e_des_0], axis=1)
    tosel = complete_cases(X)
    X_na = X.loc[tosel, ]

    e_res_na = X_na.iloc[:, 0]
    e_des_0_na = X_na.iloc[:, 1:]

    # Proceed cleaning missing data in the post-treatment period
    tosel = complete_cases(P)
    P_na = P.loc[tosel, ]

    ############################################################
    # Augment H with powers and interactions of B (not of C!!!)
    # (PolynomialFeatures is not designed to handle nans)
    u_des_0_na, e_des_0_na, e_des_1 = createPoly(u_order, e_order, index_w, u_des_0_na, e_des_0_na, e_des_1, out_feat)

    ###########################################################################
    # Estimate E[u|H], V[u|H], and Sigma

    # If the model is thought to be misspecified then E[u|H] is estimated
    if u_missp is True:
        u_des_0_na = DUflexGet(u_des_0_na, C)
        if len(u_des_0_na) <= len(u_des_0_na.columns):
            warnings.warn("Consider specifying a less complicated model for u. The number of observations used " +
                          "to parametrically predict moments is smaller than the number of covariates used." +
                          " Consider reducing either the number " +
                          "of lags (u_lags) or the order of the polynomial (u_order)!")
        u_mean = LinearRegression().fit(u_des_0_na, res_na).predict(u_des_0_na)

    elif u_missp is False:
        u_mean = 0

    # Use HC inference to estimate V[u|H]
    df = df_EST(w_constr=w_constr, w=w, A=A, B=B, J=J, KM=KM)

    Sigma, Omega = u_sigma_est(u_mean=u_mean, u_sigma=u_sigma,
                               res=res_na, Z=Z_na, V=V_na,
                               index=index, TT=TT, M=M, df=df)

    Sigma_root = sqrtm(Sigma).real

    # Auxiliary logical values to estimate bounds for w
    w_lb_est = True
    w_ub_est = True

    if w_bounds is not None:
        if sum(numpy.isnan(w_bounds[0])) == len(Y_post_fit):
            w_lb_est = False
        if sum(numpy.isnan(w_bounds[1])) == len(Y_post_fit):
            w_ub_est = False

    # Define constrained problem to be simulated
    if w_lb_est is True or w_ub_est is True:
        Q = numpy.array(Z_na).T.dot(V_na).dot(numpy.array(Z_na)) / TT
        b_arr = numpy.array(beta).flatten()

        lb = w_constr_inf['lb']
        dire = w_constr_inf['dir']
        p = w_constr_inf['p']
        QQ = w_constr_inf['Q']

        if p == "no norm":
            p_int = 0
        if p == "L1":
            p_int = 1
        if p == "L2":
            p_int = 2

        vsig = scpi_in(sims, b_arr, Sigma_root, Q, P_na, J, KM, w_lb_est,
                       w_ub_est, p, p_int, QQ, dire, lb, cores, opt_dict_inf, pass_stata)

    if w_lb_est is True:
        w_lb = numpy.nanquantile(vsig[:, :len(P_na)], q=u_alpha / 2, axis=0)
        w_lb = pandas.DataFrame(w_lb,
                                index=Y_post_fit.index,
                                columns=Y_post_fit.columns)
        fail_lb = 100 * numpy.sum(numpy.isnan(vsig[:, :len(P_na)]), axis=0) / sims

    else:
        w_lb = w_bounds[0]
        fail_lb = numpy.array([0] * len(w_bounds[0]))

    if w_ub_est is True:
        w_ub = numpy.nanquantile(vsig[:, len(P_na):], q=(1 - u_alpha / 2), axis=0)
        w_ub = pandas.DataFrame(w_ub,
                                index=Y_post_fit.index,
                                columns=Y_post_fit.columns)
        fail_ub = 100 * numpy.sum(numpy.isnan(vsig[:, len(P_na):]), axis=0) / sims
    else:
        w_ub = w_bounds[1]
        fail_ub = numpy.array([0] * len(w_bounds[1]))

    fail = numpy.c_[fail_lb, fail_ub]
    fail = numpy.array([fail_lb, fail_ub])
    failed_sims = pandas.DataFrame(fail,
                                   index=['lb', 'ub'])

    if fail.sum() > 0.1 * sims * vsig.shape[1]:
        warnings.warn("For some of the simulations used to quantify in-sample uncertainty the solution of " +
                      "the optimization problem was not found! We suggest inspecting the magnitude of this issue " +
                      "by consulting the percentage of simulations that failed contained in " +
                      "YOUR_SCPI_OBJECT_NAME.failed_sims." +
                      "In case the number of unsuccessful simulations is high, you might want to consider " +
                      "changing the stopping criteria of the algorithm through the option 'opt_list_inf'.")

    # PIs for w
    sc_l_0 = Y_post_fit + w_lb        # Left bound
    sc_r_0 = Y_post_fit + w_ub        # Left bound
    len_0 = sc_r_0 - sc_l_0           # Length

    ######################################
    ######################################
    # Estimate out-of-sample uncertainty
    ######################################
    ######################################

    sc_l_1 = sc_l_2 = sc_l_3 = sc_r_1 = sc_r_2 = sc_r_3 = None
    e_mean = e_var = numpy.nan
    # Auxiliary logical values to estimate bounds for e
    e_lb_est = True
    e_ub_est = True

    if e_bounds is not None:
        if sum(numpy.isnan(e_bounds[0])) == len(Y_post_fit):
            e_lb_est = False
        if sum(numpy.isnan(e_bounds[1])) == len(Y_post_fit):
            e_ub_est = False

    if e_method == 'gaussian' or e_method == 'all':
        e_lb, e_ub, e_1, e_2 = scpi_out(y=e_res_na, x=e_des_0_na, preds=e_des_1,
                                        e_method="gaussian", alpha=e_alpha / 2,
                                        e_lb_est=e_lb_est, e_ub_est=e_ub_est)

        # Overwrite with user's input
        if e_lb_est is False:
            e_lb = e_bounds[0]
        if e_ub_est is False:
            e_ub = e_bounds[1]

        sc_l_1 = sc_l_0 + e_lb
        sc_r_1 = sc_r_0 + e_ub
        len_1 = sc_r_1 - sc_l_1

        e_mean = e_1
        e_var = e_2

    if e_method == 'ls' or e_method == 'all':
        e_lb, e_ub, e_1, e_2 = scpi_out(y=e_res_na, x=e_des_0_na, preds=e_des_1,
                                        e_method="ls", alpha=e_alpha / 2,
                                        e_lb_est=e_lb_est, e_ub_est=e_ub_est)

        # Overwrite with user's input
        if e_lb_est is False:
            e_lb = e_bounds[0]
        if e_ub_est is False:
            e_ub = e_bounds[1]

        sc_l_2 = sc_l_0 + e_lb
        sc_r_2 = sc_r_0 + e_ub
        len_2 = sc_r_2 - sc_l_2

    if e_method == 'qreg' or e_method == 'all':
        if e_order == 0:
            e_lb = numpy.quantile(e_res_na, q=e_alpha / 2)
            e_ub = numpy.quantile(e_res_na, q=1 - e_alpha / 2)
        else:
            e_lb, e_ub, e_1, e_2 = scpi_out(y=e_res_na, x=e_des_0_na, preds=e_des_1,
                                            e_method="qreg", alpha=e_alpha / 2,
                                            e_lb_est=e_lb_est, e_ub_est=e_ub_est)
        # Overwrite with user's input
        if e_lb_est is False:
            e_lb = e_bounds[0]
        if e_ub_est is False:
            e_ub = e_bounds[1]

        sc_l_3 = sc_l_0 + e_lb
        sc_r_3 = sc_r_0 + e_ub
        len_3 = sc_r_3 - sc_l_3

    ###############################################
    # Return objects
    CI_0 = pandas.concat([sc_l_0, sc_r_0, len_0], axis=1)
    CI_0.columns = ['Left Bound', 'Right Bound', 'Length']
    CI_0.index.rename('Time', inplace=True)

    if sc_l_1 is not None:
        CI_1 = pandas.concat([sc_l_1, sc_r_1, len_1], axis=1)
        CI_1.columns = ['Left Bound', 'Right Bound', 'Length']
        CI_1.index.rename('Time', inplace=True)
    else:
        CI_1 = None

    if sc_l_2 is not None:
        CI_2 = pandas.concat([sc_l_2, sc_r_2, len_2], axis=1)
        CI_2.columns = ['Left Bound', 'Right Bound', 'Length']
        CI_2.index.rename('Time', inplace=True)
    else:
        CI_2 = None

    if sc_l_3 is not None:
        CI_3 = pandas.concat([sc_l_3, sc_r_3, len_3], axis=1)
        CI_3.columns = ['Left Bound', 'Right Bound', 'Length']
        CI_3.index.rename('Time', inplace=True)
    else:
        CI_3 = None

    u_user = u_design is not None  # True if user provided the design matrix for in-sample inference
    e_user = e_design is not None  # True if user provided the design matrix for out-of-sample inference

    ##################################################
    # Plot

    if plot is True:
        to_plot = scpi_output(b=sc_pred.b,
                              w=sc_pred.w,
                              r=sc_pred.r,
                              Y_pre_fit=sc_pred.Y_pre_fit,
                              Y_post_fit=sc_pred.Y_post_fit,
                              A_hat=sc_pred.A_hat,
                              res=sc_pred.res,
                              V=sc_pred.V,
                              w_constr=sc_pred.w_constr,
                              w_constr_inf=w_constr_inf,
                              A=sc_pred.A,
                              B=sc_pred.B,
                              C=sc_pred.C,
                              P=P_na,
                              Y_pre=sc_pred.Y_pre,
                              Y_post=sc_pred.Y_post,
                              Y_donors=sc_pred.Y_donors,
                              J=sc_pred.J,
                              K=sc_pred.K,
                              KM=sc_pred.KM,
                              M=sc_pred.M,
                              cointegrated_data=sc_pred.cointegrated_data,
                              period_pre=sc_pred.period_pre,
                              period_post=sc_pred.period_post,
                              T0_features=sc_pred.T0_features,
                              T1_outcome=sc_pred.T1_outcome,
                              outcome_var=sc_pred.outcome_var,
                              features=sc_pred.features,
                              glob_cons=sc_pred.glob_cons,
                              out_in_features=sc_pred.out_in_features,
                              CI_in_sample=CI_0,
                              CI_all_gaussian=CI_1,
                              CI_all_ls=CI_2,
                              CI_all_qreg=CI_3,
                              Sigma=Sigma,
                              u_mean=u_mean,
                              u_var=Omega,
                              e_mean=e_mean,
                              e_var=e_var,
                              u_missp=u_missp,
                              u_lags=u_lags,
                              u_order=u_order,
                              u_sigma=u_sigma,
                              u_user=u_user,
                              e_method=e_method,
                              e_lags=e_lags,
                              e_order=e_order,
                              e_user=e_user,
                              rho=rho,
                              u_alpha=u_alpha,
                              e_alpha=e_alpha,
                              sims=sims,
                              failed_sims=failed_sims,
                              plotres=None)

        plotres = scplot(result=to_plot)
    else:
        plotres = None

    return scpi_output(b=sc_pred.b,
                       w=sc_pred.w,
                       r=sc_pred.r,
                       Y_pre_fit=sc_pred.Y_pre_fit,
                       Y_post_fit=sc_pred.Y_post_fit,
                       A_hat=sc_pred.A_hat,
                       res=sc_pred.res,
                       V=sc_pred.V,
                       w_constr=sc_pred.w_constr,
                       w_constr_inf=w_constr_inf,
                       A=sc_pred.A,
                       B=sc_pred.B,
                       C=sc_pred.C,
                       P=P_na,
                       Y_pre=sc_pred.Y_pre,
                       Y_post=sc_pred.Y_post,
                       Y_donors=sc_pred.Y_donors,
                       J=sc_pred.J,
                       K=sc_pred.K,
                       KM=sc_pred.KM,
                       M=sc_pred.M,
                       cointegrated_data=sc_pred.cointegrated_data,
                       period_pre=sc_pred.period_pre,
                       period_post=sc_pred.period_post,
                       T0_features=sc_pred.T0_features,
                       T1_outcome=sc_pred.T1_outcome,
                       outcome_var=sc_pred.outcome_var,
                       features=sc_pred.features,
                       glob_cons=sc_pred.glob_cons,
                       out_in_features=sc_pred.out_in_features,
                       CI_in_sample=CI_0,
                       CI_all_gaussian=CI_1,
                       CI_all_ls=CI_2,
                       CI_all_qreg=CI_3,
                       Sigma=Sigma,
                       u_mean=u_mean,
                       u_var=Omega,
                       e_mean=e_mean,
                       e_var=e_var,
                       u_missp=u_missp,
                       u_lags=u_lags,
                       u_order=u_order,
                       u_sigma=u_sigma,
                       u_user=u_user,
                       e_method=e_method,
                       e_lags=e_lags,
                       e_order=e_order,
                       e_user=e_user,
                       rho=rho,
                       u_alpha=u_alpha,
                       e_alpha=e_alpha,
                       sims=sims,
                       failed_sims=failed_sims,
                       plotres=plotres)


# Define class

class scpi_output:
    def __init__(self, b, w, r, Y_pre_fit, Y_post_fit, A_hat, res, V, w_constr, w_constr_inf,
                 A, B, C, P, Y_pre, Y_post, Y_donors, J, K, KM, M,
                 cointegrated_data, period_pre, period_post, T0_features,
                 T1_outcome, features, outcome_var, glob_cons, out_in_features,
                 CI_in_sample, CI_all_gaussian, CI_all_ls, CI_all_qreg, Sigma,
                 u_mean, u_var, e_mean, e_var, u_missp, u_lags, u_order,
                 u_sigma, u_user, e_method, e_lags, e_order, e_user,
                 rho, u_alpha, e_alpha, sims, failed_sims, plotres):

        self.b = b
        self.w = w
        self.r = r
        self.Y_pre_fit = Y_pre_fit
        self.Y_post_fit = Y_post_fit
        self.A_hat = A_hat
        self.res = res
        self.V = V
        self.w_constr = w_constr
        self.w_constr_inf = w_constr_inf
        self.A = A
        self.B = B
        self.C = C
        self.P = P
        self.Y_pre = Y_pre
        self.Y_post = Y_post
        self.Y_donors = Y_donors
        self.J = J
        self.K = K
        self.KM = KM
        self.M = M
        self.cointegrated_data = cointegrated_data
        self.period_pre = period_pre
        self.period_post = period_post
        self.T0_features = T0_features
        self.T1_outcome = T1_outcome
        self.features = features
        self.outcome_var = outcome_var
        self.glob_cons = glob_cons
        self.out_in_features = out_in_features
        self.CI_in_sample = CI_in_sample
        self.CI_all_gaussian = CI_all_gaussian
        self.CI_all_ls = CI_all_ls
        self.CI_all_qreg = CI_all_qreg
        self.Sigma = Sigma
        self.u_mean = u_mean
        self.u_var = u_var
        self.e_mean = e_mean
        self.e_var = e_var
        self.u_missp = u_missp
        self.u_lags = u_lags
        self.u_order = u_order
        self.u_sigma = u_sigma
        self.u_user = u_user
        self.e_method = e_method
        self.e_lags = e_lags
        self.e_order = e_order
        self.e_user = e_user
        self.rho = rho
        self.u_alpha = u_alpha
        self.e_alpha = e_alpha
        self.sims = sims
        self.failed_sims = failed_sims
        self.plotres = plotres

    def __repr__(self):

        # Prepare objects to print (estimation)
        fw = 30
        if self.M == 1:
            fw_r = 14
        if self.M > 1:
            fw_r = 40

        constr = self.w_constr['name']
        if self.w_constr['Q'] is not None:
            Qsize = round(self.w_constr['Q'], 3)
        else:
            Qsize = "-"
        tr_unit = self.A.columns.values.tolist()
        pt_in = self.period_pre[0]
        pt_fi = self.period_pre[len(self.period_pre) - 1]
        ppre = str(pt_in) + '-' + str(pt_fi)

        Weights = self.w.rename(columns={0: 'Weights'}, inplace=False)
        Weights = round(Weights, 3)
        activew = len(Weights.loc[abs(Weights['Weights']) > 0])

        if self.KM > 0:
            Covariates = self.r.rename(columns={0: 'Covariates'}, inplace=False)
            Covariates = round(Covariates, 3)

        if constr is None:
            constr = "User Provided"

        # Print stuff
        print('-----------------------------------------------------------------------')
        print('Call: scpi')
        print('Synthetic Control Estimation - Setup')
        print('')

        print('Constraint Type:'.ljust(fw), str(constr).rjust(fw_r))
        print('Constraint Size (Q):'.ljust(fw), str(Qsize).rjust(fw_r))
        print('Treated Unit:'.ljust(fw), str(tr_unit[0]).rjust(fw_r))
        print('Size of the donor pool:'.ljust(fw), str(self.J).rjust(fw_r))
        print('Features'.ljust(fw), str(self.M).rjust(fw_r))
        print('Pre-treatment period'.ljust(fw), str(ppre).rjust(fw_r))

        if self.M == 1:
            T0 = [c for c in self.T0_features.values()]

            print('Pre-treatment periods (used):'.ljust(fw), str(T0[0]).rjust(fw_r))

            print('Adjustment Covariates:'.ljust(fw), str(self.KM).rjust(fw_r))
        else:
            T0v = [c for c in self.T0_features.values()]
            T0n = [c for c in self.T0_features]
            Kv = [k for k in self.K.values]
            Kn = [k for k in self.K.index.get_level_values(0)]

            print('Pre-treatment periods used in estimation per feature:'.ljust(fw))
            for cov in range(len(T0n)):
                toprint = str(T0n[cov]) + ' ' + str(T0v[cov])
                print(toprint.rjust(fw_r))

            print('Covariates used for adjustment per feature:'.ljust(fw))
            for cov in range(len(Kn)):
                toprint = str(Kn[cov]) + ' ' + str(Kv[cov])
                print(toprint.rjust(fw_r))

        print('')
        print('Synthetic Control Estimation - Results')
        print('')
        print('Active donors:', activew)
        print('')
        print('Coefficients:')
        print(Weights)
        if self.KM > 0:
            print('')
            print(Covariates)

        # Prepare objects to print (inference)
        e_method = self.e_method
        Y_tr_post = round(self.Y_post, 2)
        Y_tr_post = self.Y_post.astype(float).round(2)
        Y_sc_post = round(self.Y_post_fit, 2)
        Y_tr_post.columns = pandas.Index(['Treated'])
        Y_sc_post.columns = pandas.Index(['Synthetic'])

        if e_method == 'gaussian':
            CI = round(self.CI_all_gaussian.iloc[:, 0:2], 2)
        elif e_method == 'ls':
            CI = round(self.CI_all_ls.iloc[:, 0:2], 2)
        elif e_method == 'qreg':
            CI = self.CI_all_qreg.iloc[:, 0:2].astype(float).round(2)
        elif e_method == 'all':
            CI1 = round(self.CI_all_gaussian.iloc[:, 0:2], 2)
            CI2 = round(self.CI_all_ls.iloc[:, 0:2], 2)
            CI3 = self.CI_all_qreg.iloc[:, 0:2].astype(float).round(2)

        print('')
        print('-----------------------------------------------------------------------')
        print('Synthetic Control Inference - Setup')
        print('')

        if (self.u_user is False):
            print("In-sample Inference:                          ")
            print("     Misspecified model                       " + str(self.u_missp))
            print("     Order of polynomial (B)                  " + str(self.u_order))
            print("     Lags (B)                                 " + str(self.u_lags))
            print("     Variance-Covariance Estimator            " + self.u_sigma)
        else:
            print("In-sample Inference:")
            print("     User provided")

        if (self.e_user is False):
            print("Out-of-sample Inference:                          ")
            print("     Method                                   " + self.e_method)
            print("     Order of polynomial (B)                  " + str(self.e_order))
            print("     Lags (B)                                 " + str(self.e_lags))
        else:
            print("Out-of-sample Inference:")
            print("     User provided")

        if e_method == 'gaussian':
            print('   Inference with subgaussian bounds')
            dfprint = pandas.concat([Y_tr_post, Y_sc_post, CI], axis=1)
            dfprint.index.rename('Time', inplace=True)
            print(dfprint)
        elif e_method == 'ls':
            print('   Inference with location-scale model')
            dfprint = pandas.concat([Y_tr_post, Y_sc_post, CI], axis=1)
            dfprint.index.rename('Time', inplace=True)
            print(dfprint)
        elif e_method == 'qreg':
            print('   Inference with quantile regression')
            dfprint = pandas.concat([Y_tr_post, Y_sc_post, CI], axis=1)
            dfprint.index.rename('Time', inplace=True)
            print(dfprint)
        elif e_method == 'all':
            print('                              Subgaussian')
            dfprint = pandas.concat([Y_tr_post, Y_sc_post, CI1], axis=1)
            dfprint.index.rename('Time', inplace=True)
            print(dfprint)
            print('            Location Scale          Quantile Reg')
            print(pandas.concat([CI2, CI3], axis=1))

        else:
            print(pandas.concat([Y_tr_post, Y_sc_post, CI], axis=1).index.rename('Time'))

        if constr == "ridge":
            warnings.warn("The current version of the package does not take into account a small probability loss"
                          " in prediction intervals when using" +
                          "'ridge' as estimation method! An updated version will be rolled out soon!!")

        return ''



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 25 07:55:26 2021

@author: Filippo Palomba
"""
# Temporary code to suppress pandas FutureWarning
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import pandas
import numpy
from plotnine import ggplot, aes, geom_point, geom_errorbar, geom_vline, geom_line, theme, theme_bw
from plotnine import element_blank, labs, guide_legend, scale_color_manual, ggtitle


def scplot(result,
           col_dots_t=None,
           col_line_t=None,
           col_dots_s=None,
           col_line_s=None,
           x_lab=None,
           y_lab=None,
           e_out=True,
           e_method=None,
           save_data=None):

    '''
    Parameters
    ----------
    result : scest_output/scpi_output
        a class `scest_output' object, obtained by calling scest, or a class
        `scpi_output' object, obtained by calling scpi

    col_dots_t : str
        string indicating the color of the time series marker for treated unit

    col_line_t : str
        string indicating the color of the time series line for treated unit

    col_dots_s : str
        string indicating the color of the time series marker for synthetic control unit

    col_line_s : str
        string indicating the color of the time series line for synthetic control unit

    x_lab : str
        string indicating x axis title

    y_lab : str
        string indicating y axis title

    e_out : bool, default True
        a logical specifying whether out-of-sample uncertainty should be included in the plot(s).

    e_method : str, default "gaussian"
        a string specifying the type of uncertainty estimation used for out-of-sample uncertainty quantification.

    save_data : str
        a string specifying the name (and the folder) of the saved dataframe containing the processed data used to
        produce the plot. The data is saved in .csv format and the folder specified.

    Returns
    ----------
    plot : plotnine
        plotnine object that can be further modified.

    References
    ----------
    Abadie, A. (2021), “Using Synthetic Controls: Feasibility, Data Requirements, and Methodological
    Aspects,” Journal of Economic Literature, 59, 391-425.

    Cattaneo, M. D., Feng, Y., and Titiunik, R. (2021), “Prediction Intervals for Synthetic Control
    Methods,” Journal of the American Statistical Association, 116, 1865-1880.

    Cattaneo, M. D., Palomba, F., Feng, Y., and Titiunik, R. (2022), “scpi: Uncertainty Quantification for
    Synthetic Control Estimators”.

    See Also
    --------
    scdata, scest, scpi

    '''
    class_input = result.__class__.__name__

    if class_input not in ['scest_output', 'scpi_output']:
        raise Exception("The object 'result' should be the output of scest or scpi!")

    if x_lab is None:
        x_lab = 'Time'
    else:
        if not isinstance(x_lab, str):
            raise Exception("The option 'x_lab' should be a string!")

    if y_lab is None:
        y_lab = 'Outcome Variable'
    else:
        if not isinstance(y_lab, str):
            raise Exception("The option 'y_lab' should be a string!")

    period_pre = result.period_pre
    period_post = result.period_post
    T0 = period_pre[len(period_pre) - 1]

    if col_dots_t is None:
        col_dots_t = 'gray'
    if col_line_t is None:
        col_line_t = 'gray'
    if col_dots_s is None:
        col_dots_s = 'blue'
    if col_line_s is None:
        col_line_s = 'blue'

    # Store data for treated unit
    time = numpy.concatenate([period_pre, period_post])
    y_act = pandas.concat([result.Y_pre, result.Y_post]).to_numpy().flatten()
    data_points_act = pandas.DataFrame({'time': time,
                                        'y_act': y_act
                                        })

    # Store data for synthetic control unit
    y_sc_df = pandas.concat([result.Y_pre_fit, result.Y_post_fit])

    if class_input == 'scest_output':

        y_sc_na = pandas.DataFrame(numpy.array([numpy.nan] * len(time)))

        # Check if some periods have missing point estimate/missing CI
        not_miss_plot = [t in y_sc_df.index.tolist() for t in time]
        y_sc_na.loc[not_miss_plot, ] = y_sc_df.iloc[:, [0]].to_numpy()

        data_points_act = pandas.DataFrame({'time': time,
                                            'y_act': y_act
                                            })

        data_points = pandas.concat([data_points_act, y_sc_na], axis=1)
        data_points.columns = ['time', 'y_act', 'y_sc']
        data_points['tr'] = ['Treated'] * len(y_sc_na)
        data_points['sc'] = ['Synthetic Control'] * len(y_sc_na)

        plot_struc = (ggplot(data_points) +
                      theme_bw() +
                      theme(panel_grid=element_blank(),
                            legend_position=(.5, .05),
                            legend_direction='horizontal',
                            legend_box='horizontal',
                            legend_title=element_blank(),
                            subplots_adjust={'bottom': 0.2}) +
                      labs(x=x_lab, y=y_lab)
                      )

        plot = (plot_struc +
                geom_point(mapping=aes(x='time', y='y_act', color='tr'), shape='o', fill='white', na_rm=False)
                + geom_point(mapping=aes(x='time', y='y_sc', color='sc'), shape='o', na_rm=False) +
                geom_line(mapping=aes(x='time', y='y_act', color='tr'), na_rm=False) +
                geom_line(mapping=aes(x='time', y='y_sc', color='sc'), linetype='dashed', na_rm=False) +
                geom_vline(xintercept=T0, linetype='dotted') +
                scale_color_manual(name="", values=[col_dots_t, col_dots_s],
                                   labels=["Treated", "Synthetic Control"],
                                   guide=guide_legend(override_aes={'linetype': ['solid', 'dashed'],
                                                                    'shape': ['o', 'o']})))
        return plot

    elif class_input == 'scpi_output':
        if e_method is None:
            e_method = result.e_method
        elif e_method == 'all':
            e_method = 'qreg'
        else:
            if not e_method == result.e_method:
                raise Exception("Make sure that e_method is the same as the one in scpi!")

        sc_l_0 = result.CI_in_sample.iloc[:, [0]].to_numpy()
        sc_r_0 = result.CI_in_sample.iloc[:, [1]].to_numpy()

        if e_method == 'gaussian':
            sc_l_1 = result.CI_all_gaussian.iloc[:, [0]].to_numpy()
            sc_r_1 = result.CI_all_gaussian.iloc[:, [1]].to_numpy()

        if e_method == 'ls':
            sc_l_1 = result.CI_all_ls.iloc[:, [0]].to_numpy()
            sc_r_1 = result.CI_all_ls.iloc[:, [1]].to_numpy()

        if e_method == 'qreg':
            sc_l_1 = result.CI_all_qreg.iloc[:, [0]].to_numpy(dtype='float64')
            sc_r_1 = result.CI_all_qreg.iloc[:, [1]].to_numpy(dtype='float64')

        y_sc_na = pandas.DataFrame(numpy.array([numpy.nan] * len(time)))
        sc_l_0_na = pandas.DataFrame(numpy.array([numpy.nan] * len(time)))
        sc_r_0_na = pandas.DataFrame(numpy.array([numpy.nan] * len(time)))
        sc_l_1_na = pandas.DataFrame(numpy.array([numpy.nan] * len(time)))
        sc_r_1_na = pandas.DataFrame(numpy.array([numpy.nan] * len(time)))

        # Check if some periods have missing point estimate/missing CI
        not_miss_plot = [t in y_sc_df.index.tolist() for t in time]
        not_miss_ci = [t in result.CI_in_sample.index.tolist() for t in time]

        y_sc_na.loc[not_miss_plot, ] = y_sc_df.iloc[:, [0]].to_numpy()
        sc_l_0_na.loc[not_miss_ci, ] = sc_l_0
        sc_r_0_na.loc[not_miss_ci, ] = sc_r_0
        sc_l_1_na.loc[not_miss_ci, ] = sc_l_1
        sc_r_1_na.loc[not_miss_ci, ] = sc_r_1

        data_points_act = pandas.DataFrame({'time': time,
                                            'y_act': y_act
                                            })

        data_points = pandas.concat([data_points_act, y_sc_na,
                                    sc_l_0_na, sc_r_0_na,
                                    sc_l_1_na, sc_r_1_na], axis=1)
        data_points.columns = ['time', 'y_act', 'y_sc', 'lb0', 'ub0', 'lb1', 'ub1']

        data_points['tr'] = ['Treated'] * len(y_sc_na)
        data_points['sc'] = ['Synthetic Control'] * len(y_sc_na)

        plot_struc = (ggplot(data_points) +
                      theme_bw() +
                      theme(panel_grid=element_blank(),
                            legend_position=(.5, .05),
                            legend_direction='horizontal',
                            legend_box='horizontal',
                            legend_title=element_blank(),
                            subplots_adjust={'bottom': 0.2}) +
                      labs(x=x_lab, y=y_lab)
                      )

        plot_lines = (plot_struc +
                      geom_point(mapping=aes(x='time', y='y_act', color='tr'), shape='o', fill='white', na_rm=False) +
                      geom_point(mapping=aes(x='time', y='y_sc', color='sc'), shape='o', na_rm=False) +
                      geom_line(mapping=aes(x='time', y='y_act', color='tr'), na_rm=False) +
                      geom_line(mapping=aes(x='time', y='y_sc', color='sc'), linetype='dashed', na_rm=False) +
                      geom_vline(xintercept=T0, linetype='dotted') +
                      scale_color_manual(name="", values=[col_dots_t, col_dots_s],
                                         labels=["Treated", "Synthetic Control"],
                                         guide=guide_legend(override_aes={'linetype': ['solid', 'dashed'],
                                                                          'shape': ['o', 'o']})))

        if e_out is False:
            plot = plot_lines + geom_errorbar(mapping=aes(x='time', ymin='lb0', ymax='ub0', color='sc'),
                                              size=0.5, linetype='solid') + ggtitle('In-sample Uncertainty')

        if e_out is True:
            if e_method == 'gaussian':
                title_str = 'In and Out of Sample Uncertainty - Subgaussian Bounds'
                plot = plot_lines + geom_errorbar(mapping=aes(x='time', ymin='lb1', ymax='ub1', color='sc'),
                                                  size=0.5, linetype='solid') + ggtitle(title_str)

            if e_method == 'ls':
                title_str = 'In and Out of Sample Uncertainty - Location-scale Model'
                plot = plot_lines + geom_errorbar(mapping=aes(x='time', ymin='lb1', ymax='ub1', color='sc'),
                                                  size=0.5, linetype='solid') + ggtitle(title_str)

            if e_method == 'qreg':
                title_str = 'In and Out of Sample Uncertainty - Quantile Regression'
                plot = plot_lines + geom_errorbar(mapping=aes(x='time', ymin='lb1', ymax='ub1', color='sc'),
                                                  size=0.5, linetype='solid') + ggtitle(title_str)

        # Save data to reproduce plot
        if save_data is not None:
            data_name = save_data + ".csv"
            data_points.to_csv(data_name)

        return plot



# -*- coding: utf-8 -*-
"""
Created on Mon Aug 16 09:59:24 2021

@author: Filippo Palomba
"""
# Temporary code to suppress pandas FutureWarning
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import pandas
import numpy
import scipy.linalg
import warnings
from copy import deepcopy
from collections import Counter
from .funs import complete_cases


class scdata_output:
    def __init__(self, A, B, C, P, Y_pre, Y_post, Y_donors, J, K, KM, M,
                 cointegrated_data, period_pre, period_post, T0_features,
                 T1_outcome, outcome_var, features, glob_cons, out_in_features):

        self.A = A
        self.B = B
        self.C = C
        self.P = P
        self.Y_pre = Y_pre
        self.Y_post = Y_post
        self.Y_donors = Y_donors
        self.J = J
        self.K = K
        self.KM = KM
        self.M = M
        self.cointegrated_data = cointegrated_data
        self.period_pre = period_pre
        self.period_post = period_post
        self.T0_features = T0_features
        self.T1_outcome = T1_outcome
        self.outcome_var = outcome_var
        self.features = features
        self.glob_cons = glob_cons
        self.out_in_features = out_in_features


def scdata(df,
           id_var,
           time_var,
           outcome_var,
           period_pre,
           period_post,
           unit_tr,
           unit_co,
           features=None,
           cov_adj=None,
           cointegrated_data=False,
           anticipation=0,
           constant=False,
           report_missing=False):

    """
    Parameters
    ----------
     df : pandas.DataFrame
        a dataframe object containing the data to be processed

     id_var : str
        a character with the name of the variable containing units' IDs

     time_var : str
        a character with the name of the time variable. The time variable has to be numpy.int64, or one of
        pandas.Timestamp and numpy.datetime64. Input a numeric time variable is suggested when working with
        yearly data, whereas for all other frequencies either pandas.Timestamp or numpy.datetime64 types are preferred.

     outcome_var : str
        a character with the name of the outcome variable. The outcome variable has to be numeric.

     period_pre : array
        a numeric vector that identifies the pre-treatment period in time_var.

     period_post : array
         a numeric vector that identifies the post-treatment period in time_var.

     unit_tr : int
         a scalar that identifies the treated unit in id_var.

     unit_co : array
         a numeric vector that identifies the donor pool in id_var.

     features : list
         a character list containing the name of the feature variables used for estimation.
        If this option is not specified the default is features = outcome_var.

     cov_adj : list
        a list specifying the names of the covariates to be used for adjustment for each feature. If the user wants
        to specify the same set of covariates for all features, a single list should be provided. If instead a
        different set of covariates per feature has to be specified, then a list of lists should be provided. Note that
        in this latter case the number of sub-lists must be equal to the number of features. Moreover, the order of the
        sub-lists matters, in the sense that the first sub-list is interpreted as the set of covariates for the first
        feature, and so on. Finally, the user can specify 'constant' and 'trend' as covariates even if they are not
        present in the loaded dataframe.

     constant : bool, default False
         a logical which controls the inclusion of a constant term across features. The default value is False.

     cointegrated_data : bool, default False
        a logical that indicates if there is a belief that the data is cointegrated or not. The default value is False.
        See the Details section for more.

     anticipation : int
        a scalar that indicates the number of periods of potential anticipation effects. Default is 0.

     report_missing : bool, default False
        a logical which prints the location of missing values if present. The default value is False.

    Returns
    -------
    The function returns an object of class `scdata_output' containing the following objects

    A : pandas.DataFrame
        a dataframe containing pre-treatment features of the treated unit.

    B : pandas.DataFrame
        a dataframe containing pre-treatment features of the control units.

    C : pandas.DataFrame
        a dataframe containing covariates for adjustment.

    P : pandas.DataFrame
        a dataframe whose rows are the vectors used to predict the out-of-sample series for the synthetic unit.

    Y_pre : pandas.DataFrame
        a dataframe containing the pre-treatment outcome of the treated unit.

    Y_post : pandas.DataFrame
        a dataframe containing the post-treatment outcome of the treated unit.

    Y_donors : pandas.DataFrame
        a dataframe containing the pre-treatment outcome of the control units.

    J : int
        the number of control units

    K : array
        a numeric array with the number of covariates used for adjustment for each feature

    KM : int
        the total number of covariates used for adjustment

    M : int
        number of features

    period_pre : array
        a numeric array with the pre-treatment period

    period_post : array
        a numeric array with the post-treatment period

    T0_features : array
        a numeric array with the number of periods used in estimation for each feature

    T1_outcome : int
        the number of post-treatment periods

    glob_cons : bool
        for internal use only

    out_in_features : bool
        for internal use only

    References
    ----------
    Abadie, A. (2021), “Using Synthetic Controls: Feasibility, Data Requirements, and Methodological
    Aspects,” Journal of Economic Literature, 59, 391-425.

    Cattaneo, M. D., Feng, Y., and Titiunik, R. (2021), “Prediction Intervals for Synthetic Control
    Methods,” Journal of the American Statistical Association, 116, 1865-1880.

    Cattaneo, M. D., Palomba, F., Feng, Y., and Titiunik, R. (2022), “scpi: Uncertainty Quantification for
    Synthetic Control Estimators”.

    See Also
    --------
    scest, scpi, scplot

    """
    # Check main input is a dataframe
    if not isinstance(df, pandas.DataFrame):
        raise Exception('Data input should be a dataframe object!')

    data = deepcopy(df)

    # Store variable names and indexes
    var_names = data.columns
    indexes = data.index.names

    # Check inputs are strings
    if not isinstance(id_var, str):
        raise Exception("You should specify the name of id_var as a string! (eg. id_var = 'ID')")

    if not isinstance(outcome_var, str):
        raise Exception("You should specify the name of outcome_var as a string! (eg. outcome_var = 'outcome')")

    if not isinstance(time_var, str):
        raise Exception("You should specify the name of time_var as a string! (eg. time_var = 'time')")

    if features is None:
        features = [outcome_var]
    elif not isinstance(features, list):
        raise Exception("The object 'features' should be a list!")

    # Check covariates for adjustment are in dataframe
    if cov_adj is not None:
        if not isinstance(cov_adj, list):
            raise Exception("The argument cov_adj should be a list!")

        if isinstance(cov_adj[0], list):  # feature-specific cov adj
            if len(cov_adj) != len(features):
                raise Exception("When specifying covariate adjustment separately for each feature make sure " +
                                "to do it for all features! You specified covariate adjustment for " +
                                str(len(cov_adj)) + " features when you currently have " +
                                str(len(features)) + " features!")
            unique_covs = [cov for sublist in cov_adj for cov in sublist]

        else:
            unique_covs = cov_adj
        # Check all covariates other than constant and trend are in dataframe

        unique_cov = list(set(unique_covs))
        unique_covs = [x for x in unique_cov if x not in ['constant', 'trend']]

        if len(unique_covs) > 0:
            if not all(cov in var_names for cov in unique_covs):
                cov_adj_not_found = [cov for cov in unique_covs if cov in var_names]
                cov_print = ' '.join(str(cov) for cov in cov_adj_not_found)
                raise Exception("Some of the covariates in cov_adj are not in the input dataframe!"
                                + cov_print)

    # Check variables are in dataframe (id and time can be either variables or indices)
    if (id_var not in var_names) and (id_var not in indexes):
        raise Exception("ID variable (id_var) not found in the input dataframe neither as a variable nor as an index!")

    if (time_var not in var_names) and (time_var not in indexes):
        raise Exception("Time variable (time_var) not found in the input df neither as a variable nor as an index!")

    if outcome_var not in var_names:
        raise Exception("Outcome variable (outcome_var) not found in the input dataframe!")

    # Make time and id columns if these variables are indexes of the dataframe
    if id_var in indexes:
        data['__ID'] = data.index.get_level_values(id_var)
    else:
        data.rename(columns={id_var: '__ID'}, inplace=True)

    if time_var in indexes:
        data['__time'] = data.index.get_level_values(time_var)
    else:
        data.rename(columns={time_var: '__time'}, inplace=True)

    # Create index
    data.set_index(['__ID', '__time'], drop=False, inplace=True)

    dd = data.iloc[0, data.columns.get_loc('__time')]
    if not isinstance(dd, (numpy.int64, numpy.int32, numpy.int16, pandas.Timestamp, numpy.datetime64)):
        raise Exception("The object time_var should be of type int, numpy.datetime64, or pandas.Timestamp!")

    if not isinstance(period_pre, numpy.ndarray):
        raise Exception("The object period_pre should be of type numpy.ndarray (eg. use numpy.arange or numpy.array)!")

    if not isinstance(period_post, numpy.ndarray):
        raise Exception("The object period_post should be of type numpy.ndarray!")

    if not isinstance(period_pre[0], (numpy.int64, numpy.int32, numpy.int16, numpy.datetime64, pandas.Timestamp)):
        raise Exception("Elements of period_pre should either be of type int or numpy.datetime64!")

    if not isinstance(period_post[0], (numpy.int64, numpy.int32, numpy.int16, numpy.datetime64, pandas.Timestamp)):
        raise Exception("Elements of period_post should either be of type int or numpy.datetime64!")

    if not isinstance(unit_co, list):
        raise Exception("The object unit_co should be of type list!")

    if not isinstance(unit_tr, list):
        unit_tr_list = [unit_tr]
    else:
        unit_tr_list = unit_tr

    if not data[outcome_var].dtype in ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']:
        raise Exception("Outcome variable (outcome_var) must be numeric!")

    if not all(feature in var_names for feature in features):
        fe_not_found = [feature for feature in features if feature not in var_names]
        fe_print = ' '.join(str(feature) for feature in fe_not_found)
        raise Exception("The following features are not in the input dataframe:"
                        + fe_print)

    if outcome_var in features:
        out_in_features = True
    else:
        out_in_features = False

    if cov_adj is not None:
        if len(features) == 1 and constant is True and 'constant' in unique_cov:
            raise Exception("When specifying just one feature you either specify constant == True" +
                            " or include 'constant' in cov_adj!")

    # Sort period variables
    period_pre = numpy.sort(period_pre)
    period_post = numpy.sort(period_post)

    # Create ID and time variables
    idd = list(set(data['__ID'].to_list()))   # unique IDs
    time = data['__time'].unique()             # unique periods

    # Check that specified units are in dataframe
    if not all(tr in idd for tr in unit_tr_list):
        raise Exception("There is no treated unit with the specified ID (unit_tr) in the specified" +
                        "ID variable (id_var)!")

    if not all(donor in idd for donor in unit_co):
        donors_not_found = [dnr for dnr in unit_co if dnr not in idd]
        donors_print = ' '.join(str(donor) for donor in donors_not_found)

        raise Exception("The following control unit(s) are not in the input dataframe:" +
                        donors_print)

    if any(co in unit_tr_list for co in unit_co):
        raise Exception("The treated unit is also contained in the donor pool!")

    if len(unit_co) < 2:
        raise Exception("Please provide at least two control units!")

    # Check specified time periods are in dataframe
    if not all(period in time for period in period_pre):
        period_not_found = [period for period in period_pre if period not in time]
        period_print = ' '.join(str(period) for period in period_not_found)
        raise Exception("The following pre-treatment period(s) are not in the input dataframe:" +
                        period_print)

    if not all(period in time for period in period_post):
        period_not_found = [period for period in period_post if period not in time]
        period_print = ' '.join(str(period) for period in period_not_found)
        raise Exception("The following post-treatment period(s) are not in the input dataframe:" +
                        period_print)

    if any(period in period_post for period in period_pre):
        raise Exception("There is an overlap between the pre-treatment period and post-treatment period!")

    # Consider eventual anticipation effect
    if anticipation > 0:
        t0 = len(period_pre)
        d = anticipation
        period_post = numpy.insert(period_post, 0, period_pre[(t0 - d):t0])
        period_pre = period_pre[:(t0 - d)]

    ############################################################################
    # Data preparation

    # Make the panel balanced
    data_bal = deepcopy(data.unstack().stack(dropna=False))
    data_bal['__ID'] = data_bal.index.get_level_values('__ID')
    data_bal['__time'] = data_bal.index.get_level_values('__time')

    rows_tr_pre = data_bal.loc[(unit_tr_list, period_pre), ]
    rows_tr_post = data_bal.loc[(unit_tr_list, period_post), ]
    rows_co_pre = data_bal.loc[(unit_co, period_pre), ]
    rows_co_post = data_bal.loc[(unit_co, period_post), ]

    ############################################################################
    # Estimation Data

    # Actual Pre-treatment Series
    Y_pre = rows_tr_pre[[outcome_var]].reset_index(level='__ID', drop=True)
    Y_pre.columns = unit_tr_list

    if len(unit_tr_list) == 1:
        label = unit_tr_list[0]
    else:
        label = "Treated"

    # Create A
    A_df = pandas.melt(rows_tr_pre,
                       id_vars=['__ID', '__time'],
                       value_vars=features,
                       var_name='feature',
                       value_name=label)

    A = A_df.loc[:, [unit_tr]]

    # Create B

    # Stack features one on top of the other
    B_df = pandas.melt(rows_co_pre,
                       id_vars=['__ID', '__time'],
                       value_vars=features,
                       var_name='feature',
                       value_name='value')

    # make the df wide so that countries are one next to the other
    B = B_df.pivot(index=['feature', '__time'],  # index to keep fixed
                   columns='__ID',                # dimension to wide
                   values='value')               # values to be widen

    # Create list that dictates order of donors
    donor_order = B.columns.values.tolist()

    # Create matrix with outcome for the donors
    sel = rows_co_pre[[outcome_var, '__time', '__ID']]
    Y_donors = sel.pivot(index='__time', columns='__ID', values=outcome_var)
    Y_donors = Y_donors[donor_order]

    # Create C
    C = pandas.DataFrame(None)
    C_names = []

    if cov_adj is not None:

        # Fixed covariate adjustment across features
        if not isinstance(cov_adj[0], list):
            covs_adj = deepcopy(cov_adj)

            # Check that constant/time trend are required by the user
            if 'constant' in cov_adj:
                covs_adj = [cov for cov in covs_adj if not cov == 'constant']
                C['constant'] = numpy.ones(len(rows_tr_pre))

            if 'trend' in cov_adj:
                covs_adj = [cov for cov in covs_adj if not cov == 'trend']
                C['trend'] = period_pre - period_pre[0] + 1

            # Add other covariates from dataframe
            rows_C = data_bal.loc[(unit_co[0], period_pre), covs_adj].reset_index()
            C[covs_adj] = rows_C[covs_adj]

            for num in range(1, len(features) + 1):
                for cov in C.columns:
                    C_names.append(str(num) + '_' + cov)

            # Create block diagonal matrix
            C_arr = numpy.kron(numpy.identity(len(features)), numpy.array(C))
            C = pandas.DataFrame(data=C_arr, columns=C_names)

        else:
            C_names = []
            for m in range(len(features)):
                C_m = pandas.DataFrame(None)
                covs_adj = cov_adj[m]

                # Check that constant/time trend are required by the user
                if 'constant' in covs_adj:
                    covs_adj = [cov for cov in covs_adj if not cov == 'constant']
                    C_m['constant'] = numpy.ones(len(rows_tr_pre))

                if 'trend' in covs_adj:
                    covs_adj = [cov for cov in covs_adj if not cov == 'trend']
                    C_m['trend'] = period_pre - period_pre[0] + 1

                # Add other covariates from dataframe
                rows_C = data_bal.loc[(unit_co[0], period_pre), covs_adj].reset_index()
                C_m[covs_adj] = rows_C[covs_adj]

                if m == 0:
                    C = C_m
                else:
                    C = scipy.linalg.block_diag(numpy.array(C), numpy.array(C_m))

                feat_num = m + 1

                for cov in cov_adj[m]:
                    C_names.append(str(feat_num) + '_' + cov)

            C = pandas.DataFrame(data=C, columns=C_names)

            # take into account empty covariate adjustemnt
            for m in range(len(features)):
                if len(cov_adj[m]) == 0:
                    i_start = len(rows_tr_pre) * m - 1
                    zeros = pandas.DataFrame(numpy.zeros((len(rows_tr_pre), len(C.columns))),
                                             columns=C_names)
                    if i_start < 0:
                        C_up = pandas.DataFrame(None)
                    else:
                        C_up = C.iloc[:i_start]

                    C_bo = C.iloc[(i_start + 1):]
                    C = pandas.concat([C_up, zeros, C_bo]).reset_index(drop=True)

    if constant is True:
        glob_c = pandas.DataFrame(data=numpy.ones(len(B)),
                                  columns=['0_constant'])
        if len(C.columns) == 0:
            C = glob_c
        else:
            C.insert(0, '0_constant', glob_c)

        C_names.insert(0, '0_constant')

    ##############################################################################
    # Prediction Data

    # Actual post-treatment series
    Y_post = rows_tr_post[[outcome_var]].reset_index(level='__ID', drop=True)
    Y_post.columns = unit_tr_list

    # Prediction matrix
    P = rows_co_post[[outcome_var, '__ID', '__time']].pivot(index='__time',
                                                            columns='__ID',
                                                            values=outcome_var)

    P = P[donor_order]

    # If the outcome variable is within the specified features then we need to
    # augment P with the corresponding (eventual) covariates used for adjustment,
    # If instead the outcome variable is not within the specified features
    # P is composed by the outcome variable of the donors only

    if out_in_features is True:
        # Check that global constant is required by the user
        if constant is True:
            P['0_constant'] = numpy.ones(len(rows_tr_post))

        # Add covariates used for adjustment in outcome variable equation (if present)
        if cov_adj is not None:

            if not isinstance(cov_adj[0], list):
                covs_adj = deepcopy(cov_adj)
                if 'constant' in covs_adj:
                    const = numpy.ones(len(rows_tr_post))
                    const_n = str(1) + '_constant'
                    P.insert(len(P.columns), const_n, const)
                    covs_adj.remove('constant')

                if 'trend' in covs_adj:
                    trend = period_post - period_pre[0] + 1
                    trend_n = str(1) + '_trend'
                    P.insert(len(P.columns), trend_n, trend)
                    covs_adj.remove('trend')

                # Add other covariates from dataframe
                rows_P = data_bal.loc[(unit_co[0], period_post), covs_adj].reset_index()
                rows_P.set_index(P.index, inplace=True)
                coln = [str(1) + '_' + cov_n for cov_n in covs_adj]
                P[coln] = rows_P[covs_adj]

            else:
                m = 1
                for covs_adj in cov_adj:
                    if m == 1:
                        if 'constant' in covs_adj:
                            const = numpy.ones(len(rows_tr_post))
                            const_n = str(m) + '_constant'
                            P.insert(len(P.columns), const_n, const)
                            covs_adj.remove('constant')

                        if 'trend' in covs_adj:
                            trend = period_post - period_pre[0] + 1
                            trend_n = str(m) + '_trend'
                            P.insert(len(P.columns), trend_n, trend)
                            covs_adj.remove('trend')

                        # Add other covariates from dataframe
                        rows_P = data_bal.loc[(unit_co[0], period_post), covs_adj].reset_index()
                        rows_P.set_index(P.index, inplace=True)
                        coln = [str(m) + '_' + cov_n for cov_n in covs_adj]
                        P[coln] = rows_P[covs_adj]
                    else:
                        coln = [str(m) + '_' + cov_n for cov_n in covs_adj]
                        zeros = numpy.zeros((len(P.index), len(covs_adj)))
                        P[coln] = zeros

                    m = m + 1

    T1 = len(period_post)

    ############################################################################
    # Proceed cleaning missing data in the pre-treatment period

    A.set_index(B.index, drop=True, inplace=True)
    if len(C.columns) > 0:
        C.set_index(B.index, drop=True, inplace=True)

    X = pandas.concat([A, B, C], axis=1)
    X_na = X.loc[complete_cases(X), ]

    A_na = X_na.loc[:, [unit_tr]]
    B_na = X_na[donor_order]
    C_na = pandas.DataFrame(None)
    if len(C.columns) == 1:
        C_na = X_na.loc[:, C_names]
    elif len(C.columns) > 1:
        C_na = X_na.loc[:, C_names]

    # Store effective number of observations per feature
    T0_features = Counter(A_na.index.get_level_values('feature'))

    ############################################################################
    # Proceed cleaning missing data in the post-treatment period
    # P_na = P.loc[complete_cases(P), ]

    ############################################################################
    # Throw warnings for missing values
    # Report missing values in pre-treatment period

    if report_missing is True and len(A) != len(A_na):
        warnings.warn('Missing values detected in the pre-treatment period!')

        # Report missing values in A
        A_missing = complete_cases(A) is False  # identify missing rows
        if numpy.sum(A_missing) > 0:
            A_rows = A.loc[A_missing, ].index.to_frame()
            A_rows.columns = ['Feature', 'Time']
            print('Missing values detected in the following feature(s) of the treated unit:')
            print(A_rows.to_string(index=False))
        else:
            print('The feature(s) of the treated unit do not contain missing values')

        # Report missing values in B
        B_missing = complete_cases(B) is False  # identify missing rows
        if numpy.sum(B_missing) > 0:
            B_rows = B.loc[B_missing, ].index.to_frame()
            B_rows.columns = ['Feature', 'Time']
            print('Missing values detected in the following feature(s) of the donor pool:')
            print(B_rows.to_string(index=False))
        else:
            print('The feature(s) of the control units do not contain missing values')

        # Report missing values in C
        if C is not None:
            C_missing = complete_cases(C) is False  # identify missing rows
            if numpy.sum(C_missing) > 0:
                C_rows = C.loc[C_missing, ].index.to_frame()
                C_rows.columns = ['Feature', 'Time']
                print('Missing values detected in the following feature(s) of the treated unit:')
                print(C_rows.to_string(index=False))
            else:
                print('The covariate(s) used for adjustment do not contain missing values.')

    # elif report_missing == False and len(A) != len(A_na):
    #     warnings.warn('Missing values detected in the pre-treatment period!' +
    #                   'If you want to see where the missing values are located re-run the command' +
    #                   'with the option report_missing = True.')
    #     suggest = 0

    # Report missing values in post-treatment period
    P_missing = complete_cases(P) is False
    if numpy.sum(P_missing) > 0:
        warnings.warn('Missing values detected in the post-treatment period! Point estimate and prediction interval' +
                      'will not be computed for some of the required periods!')

        if report_missing is True:
            P_rows = P.loc[P_missing, ].index.to_frame()
            P_rows.columns = ['Feature', 'Time']
            print('Missing values detected in the data for post-treatment prediction in the following periods:')
            print(P_rows.to_string(index=False))
        # elif suggest == 1:
        #     print('If you want to see where the missing values are located re-run the command' +
        #               'with the option report_missing = True.')

    ############################################################################
    # Store objects

    # Size of donor pool
    J = len(unit_co)

    # Total number of covariates used for adjustment
    if C is not None:
        KM = len(C.columns)
    else:
        KM = 0

    # Numer of features
    M = len(features)

    # Array containing number of covariates used for adjustment in each equation

    if cov_adj is None:
        K = pandas.DataFrame(numpy.zeros(M), index=features)
    elif not isinstance(cov_adj[0], list):
        K = pandas.DataFrame([len(cov_adj)])
    elif isinstance(cov_adj[0], list):
        k = [len(cov) for cov in cov_adj]
        K = pandas.DataFrame(k)

    if constant is True:
        K = K + 1

    K.reset_index(inplace=True)
    K.columns = ['Feature', 'Num of Covariates']
    K = K.astype({'Feature': str, 'Num of Covariates': int})

    return scdata_output(A=A_na, B=B_na, C=C_na, P=P, Y_pre=Y_pre,
                         Y_post=Y_post, Y_donors=Y_donors, J=J, K=K,
                         KM=KM, M=M, cointegrated_data=cointegrated_data,
                         period_pre=period_pre, period_post=period_post,
                         T0_features=T0_features, T1_outcome=T1,
                         outcome_var=outcome_var, features=features,
                         glob_cons=constant, out_in_features=out_in_features)
